"""BasketCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class BasketCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Basket")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Baskets"

    @property
    def commentary(self) -> str:
        return (
            "The number of baskets. Typically the dataframe supplied "
            + "to this feature will have many transactions for the same basket, "
            + "this feature allows you to determine how many shopping baskets "
            + "existed in that collection of transactions."
        )



"""Test FakeTransactions"""

from pyspark.sql import DataFrame
import pyspark.sql.functions as f
from jstark.sample.transactions import FakeGroceryTransactions


def test_fake_transactions_returns_a_dataframe():
    """In order to get 100% code coverage we need to call
    FakeGroceryTransactions().get_df() without a seed
    """
    assert isinstance(FakeGroceryTransactions().df, DataFrame)


def test_number_of_baskets_is_correct():
    """
    Had a bug where FakeGroceryTransactions was returning the wrong number of baskets.
    First wrote this test to verify the correct behaviour and then made the test pass.
    Might as well leave the test here.
    """
    number_of_baskets = 1234
    first = (
        FakeGroceryTransactions(number_of_baskets=number_of_baskets)
        .df.groupBy()
        .agg(f.countDistinct("Basket").alias("baskets"))
        .first()
    )
    assert first is not None
    assert first["baskets"] == number_of_baskets



"""Count feature"""

from typing import Callable
import pyspark.sql.functions as f
from pyspark.sql import Column


from jstark.features.feature import BaseFeature


class Count(BaseFeature):
    def aggregator(self) -> Callable[[Column], Column]:
        return self.count_aggregator

    def column_expression(self) -> Column:
        return f.lit(1)

    def default_value(self) -> Column:
        return f.lit(0)

    @property
    def description_subject(self) -> str:
        return "Count of rows"



"""ChannelCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class ChannelCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Channel")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Channels"

    @property
    def commentary(self) -> str:
        return (
            "Channels may be things like 'instore' or 'online', although you can supply"
            + " whatever values you want. Perhaps they are marketing channels. This "
            + " feature is the number of distinct channels in which activity occurred."
        )



"""RecencyWeightedBasket feature"""

from datetime import date

from jstark.features.feature import DerivedFeature

from pyspark.sql import Column
import pyspark.sql.functions as f

from jstark.feature_period import FeaturePeriod
from .basket_count import BasketCount
from .approx_basket_count import ApproxBasketCount


class RecencyWeightedApproxBasket(DerivedFeature):
    """RecencyWeightedApproxBasket"""

    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        smoothing_factor: float,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(as_at, feature_period, first_day_of_week, use_absolute_periods)
        self.__smoothing_factor = smoothing_factor

    @property
    def smoothing_factor(self) -> float:
        return self.__smoothing_factor

    def column_expression(self) -> Column:
        expr = f.lit(0.0)
        for period in range(self.feature_period.end, self.feature_period.start + 1):
            expr = expr + ApproxBasketCount(
                as_at=self.as_at,
                feature_period=FeaturePeriod(
                    self.feature_period.period_unit_of_measure, period, period
                ),
                first_day_of_week=self._first_day_of_week,
            ).column * pow(self.smoothing_factor, period)
        return expr

    def default_value(self) -> Column:
        return f.lit(None)

    @property
    def description_subject(self) -> str:
        return (
            "Exponentially weighted moving average, with smoothing factor of"
            + f" {self.smoothing_factor}, of the approximate number of baskets"
            + f" per {self.feature_period.period_unit_of_measure.name.lower()}"
        )

    @property
    def feature_name(self) -> str:
        return (
            "RecencyWeightedApproxBasket"
            + f"{self.feature_period.period_unit_of_measure.name.title()}s"
            + f"{int(self.smoothing_factor * 100)}"
            + f"_{self.feature_period.mnemonic}"
        )

    @property
    def commentary(self) -> str:
        return (
            "Exponential smoothing "
            + "(https://en.wikipedia.org/wiki/Exponential_smoothing)"
            + " is an alternative to a simple moving average which"
            + " gives greater weighting to more recent observations, thus is an"
            + " exponentially weighted moving average. It uses a smoothing factor"
            + f" between 0 & 1 which for this feature is {self.smoothing_factor}."
            + " Here the approximate number of baskets per"
            + f" {self.feature_period.period_unit_of_measure.name.lower()} is smoothed."
            + " This feature is considered to be a highly effective predictor of future"
            + " purchases, if a customer has bought a product recently then there's a"
            + " relatively high probability they will buy it again."
            + f" This is less accurate than {self.feature_name.replace('Approx', '')}"
            + " though is less computationally expensive to calculate because it "
            + " does not calculate a distinct count for each"
            + f" {self.feature_period.period_unit_of_measure.name.lower()}."
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(as_at={self.as_at}"
            f", feature_period='{self.feature_period.mnemonic}'"
            f", smoothing_factor={self.smoothing_factor}"
            f", first_day_of_week={self._first_day_of_week!r})"
        )


class RecencyWeightedBasket(RecencyWeightedApproxBasket):
    """RecencyWeightedBasket feature"""

    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        smoothing_factor: float,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at,
            feature_period,
            smoothing_factor,
            first_day_of_week,
            use_absolute_periods,
        )

    def column_expression(self) -> Column:
        expr = f.lit(0.0)
        for period in range(self.feature_period.end, self.feature_period.start + 1):
            expr = expr + BasketCount(
                as_at=self.as_at,
                feature_period=FeaturePeriod(
                    self.feature_period.period_unit_of_measure, period, period
                ),
                first_day_of_week=self._first_day_of_week,
            ).column * pow(super().smoothing_factor, period)
        return expr

    @property
    def description_subject(self) -> str:
        """simply RecencyWeightedApproxBasketXX_periodmenmonic's description with the
        word approximate removed"""
        return super().description_subject.replace("approximate ", "")

    @property
    def commentary(self) -> str:
        return (
            "Exponential smoothing "
            + "(https://en.wikipedia.org/wiki/Exponential_smoothing)"
            + " is an alternative to a simple moving average which"
            + " gives greater weighting to more recent observations, thus is an"
            + " exponentially weighted moving average. It uses a smoothing factor"
            + f" between 0 & 1 which for this feature is {self.smoothing_factor}."
            + " Here the number of baskets per"
            + f" {self.feature_period.period_unit_of_measure.name.lower()} is smoothed."
            + " This feature is considered to be a highly effective predictor of future"
            + " purchases, if a customer has bought a product recently then there's a"
            + " relatively high probability they will buy it again."
            + " This is computationally expensive to calculate because it "
            + " requires a distinct count of baskets for each"
            + f" {self.feature_period.period_unit_of_measure.name.lower()}. Every"
            + " distinct count operation is expensive so the less that are performed,"
            + " the better (YMMV based on a number of factors, "
            + "mainly the volume of data"
            + " being processed). For this reason you should consider choosing a small"
            + f" number of {self.feature_period.period_unit_of_measure.name.lower()}s"
            + " for the feature period. This feature"
            + f" ({self.feature_name}) is for"
            + f" {self.feature_period.number_of_periods}"
            + f" {self.feature_period.period_unit_of_measure.name.lower()}"
            + f"{'s' if self.feature_period.number_of_periods > 1 else ''}. You might"
            + f" consider using {self.feature_name.replace('Basket', 'ApproxBasket')}"
            + " instead which is less accurate but computationally cheaper."
        )

    @property
    def feature_name(self) -> str:
        """simply RecencyWeightedApproxBasketXX_periodmenmonic's with the
        word approximate removed"""
        return super().feature_name.replace("Approx", "")


class RecencyWeightedBasket90(RecencyWeightedBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.9, first_day_of_week, use_absolute_periods
        )


class RecencyWeightedBasket95(RecencyWeightedBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.95, first_day_of_week, use_absolute_periods
        )


class RecencyWeightedBasket99(RecencyWeightedBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.99, first_day_of_week, use_absolute_periods
        )


class RecencyWeightedApproxBasket90(RecencyWeightedApproxBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.9, first_day_of_week, use_absolute_periods
        )


class RecencyWeightedApproxBasket95(RecencyWeightedApproxBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.95, first_day_of_week, use_absolute_periods
        )


class RecencyWeightedApproxBasket99(RecencyWeightedApproxBasket):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at, feature_period, 0.99, first_day_of_week, use_absolute_periods
        )



from datetime import datetime, date
import pytest
from pyspark.sql import DataFrame, SparkSession

from jstark.mealkit.mealkit_features import MealkitFeatures
from jstark.features.feature import FeaturePeriod, PeriodUnitOfMeasure


def test_orderweeks(
    as_at_timestamp: datetime, dataframe_of_faker_mealkit_orders: DataFrame
):
    """Test OrderWeeks

    OrderWeeks is the number of weeks in which at least one order was placed

    """
    mf = MealkitFeatures(as_at=as_at_timestamp, feature_periods=["52w0"])
    output_df = (
        dataframe_of_faker_mealkit_orders.groupBy()
        .agg(*mf.features)
        .select("OrderCount_52w0")
    )
    first = output_df.first()
    assert first is not None
    assert first["OrderCount_52w0"] == 903


def test_as_at_timestamp(dataframe_of_faker_mealkit_orders: DataFrame):
    """
    We had a situation where as_at was being passed as a datetime (which we support)
    however an error was still occurring:
    > TypeError: can't compare datetime.datetime to datetime.date
    Following TDD this was written as a failing test which was then fixed,
    the fix is in the same commit as this test.
    """
    as_at_timestamp = datetime(2022, 11, 30, 10, 12, 13)
    mf = MealkitFeatures(
        as_at=as_at_timestamp, feature_periods=["1q1", "2q2", "3q3", "4q4"]
    )
    dataframe_of_faker_mealkit_orders.groupBy().agg(*mf.features)


def test_desired_features(dataframe_of_faker_mealkit_orders: DataFrame):
    mf = MealkitFeatures(
        as_at=date(2022, 11, 30),
        feature_periods=["1q1", "2q2", "3q3", "4q4"],
        feature_stems=["OrderCount"],
    )
    output_df = dataframe_of_faker_mealkit_orders.groupBy().agg(*mf.features)
    assert output_df.columns == [
        "OrderCount_1q1",
        "OrderCount_2q2",
        "OrderCount_3q3",
        "OrderCount_4q4",
    ]


def test_non_existent_desired_features(dataframe_of_faker_mealkit_orders: DataFrame):
    mf = MealkitFeatures(
        as_at=date(2022, 11, 30),
        feature_periods=["1q1", "2q2", "3q3", "4q4"],
        feature_stems=["NonExistentFeature", "NonExistentFeature2"],
    )
    with pytest.raises(Exception) as exc_info:
        dataframe_of_faker_mealkit_orders.groupBy().agg(*mf.features)
    assert (
        str(exc_info.value)
        == "Feature(s) ['NonExistentFeature', 'NonExistentFeature2'] not found"
    )


def test_cuisines(dataframe_of_faker_mealkit_orders: DataFrame):
    mf = MealkitFeatures(
        as_at=date(2022, 1, 1),
        feature_periods=["1q1", "2q2", "3q3", "4q4"],
        feature_stems=["Cuisines"],
    )
    output_df = dataframe_of_faker_mealkit_orders.groupBy().agg(*mf.features)
    assert output_df.columns == [
        "Cuisines_1q1",
        "Cuisines_2q2",
        "Cuisines_3q3",
        "Cuisines_4q4",
    ]
    first = output_df.first()
    assert first is not None
    assert first["Cuisines_1q1"] == ["Italian", "French", "Spanish"]
    assert first["Cuisines_2q2"] == ["Italian", "French", "Spanish"]
    assert first["Cuisines_3q3"] == ["Italian", "French", "Spanish"]
    assert first["Cuisines_4q4"] == ["Italian", "French", "Spanish"]


def test_cuisine(spark_session: SparkSession):
    df = spark_session.createDataFrame(
        [
            ("Italian", datetime(2022, 1, 10, 1, 2, 3)),
            ("Italian", datetime(2022, 1, 10, 1, 2, 3)),
            ("Italian", datetime(2022, 1, 10, 1, 2, 3)),
            ("French", datetime(2022, 1, 10, 1, 2, 3)),
            ("French", datetime(2022, 1, 10, 1, 2, 3)),
            ("Spanish", datetime(2022, 1, 10, 1, 2, 3)),
        ],
        ["cuisine", "timestamp"],
    )
    mf = MealkitFeatures(
        as_at=date(2022, 2, 1),
        feature_periods=["1m1"],
        feature_stems=[
            "ItalianCuisineCount",
            "FrenchCuisineCount",
            "SpanishCuisineCount",
        ],
    )
    output_df = df.groupBy().agg(*mf.features)
    assert (
        output_df.schema["ItalianCuisineCount_1m1"].metadata["description"]
        == "Count of Italian recipes between 2022-01-01 and 2022-01-31"
    )
    assert output_df.schema["ItalianCuisineCount_1m1"].metadata["commentary"] == (
        "The number of Italian recipes. Typically the dataframe supplied to this "
        + "feature will have many recipes for the same cuisine, this feature allows "
        + "you to determine how many Italian recipes have been ordered."
    )
    first = output_df.first()
    assert first is not None
    assert first["ItalianCuisineCount_1m1"] == 3
    assert first["FrenchCuisineCount_1m1"] == 2
    assert first["SpanishCuisineCount_1m1"] == 1
    assert [
        c.metadata["description"]
        for c in output_df.schema
        if c.name == "ItalianCuisineCount_1m1"
    ][0] == "Count of Italian recipes between 2022-01-01 and 2022-01-31"


def test_none_args(spark_session: SparkSession):
    mf = MealkitFeatures(as_at=date(2022, 2, 1))
    assert mf.feature_periods == [FeaturePeriod(PeriodUnitOfMeasure.WEEK, 52, 0)]
    assert mf.feature_stems == set[str]()



"""Base class for all feature generators"""

from abc import ABCMeta
import re
from datetime import date
from typing import Self

from pyspark.sql import Column, SparkSession

from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure
from jstark.exceptions import FeaturePeriodMnemonicIsInvalid
from jstark.features.feature import Feature

FeaturePeriodsType = (
    list[FeaturePeriod]
    | list[str]
    | list[FeaturePeriod | str]
    | set[FeaturePeriod]
    | set[str]
    | None
)


class FeatureGenerator(metaclass=ABCMeta):
    """Base class for all feature generators"""

    def __init__(
        self,
        as_at: date | None = None,
        feature_periods: FeaturePeriodsType = None,
        feature_stems: set[str] | list[str] | None = None,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        if as_at is None:
            as_at = date.today()
        if feature_stems is None:
            feature_stems = set[str]()
        if isinstance(feature_stems, list):
            feature_stems = set[str](feature_stems)
        self.as_at = as_at
        self.feature_periods = feature_periods
        self.feature_stems = feature_stems
        self.first_day_of_week = first_day_of_week
        self.use_absolute_periods = use_absolute_periods

    FEATURE_CLASSES: set[type["Feature"]] = set[type["Feature"]]()

    @property
    def as_at(self) -> date:
        return self._as_at

    @as_at.setter
    def as_at(self, value: date) -> None:
        self._as_at = value

    @property
    def feature_periods(self) -> list[FeaturePeriod]:
        return self.__feature_periods

    @feature_periods.setter
    def feature_periods(self, value: FeaturePeriodsType) -> None:
        if value is None:
            value = {FeaturePeriod(PeriodUnitOfMeasure.WEEK, 52, 0)}
        period_unit_of_measure_values = "".join([e.value for e in PeriodUnitOfMeasure])
        regex = (
            # https://regex101.com/r/Xvf3ey/1
            r"^(\d*)([" + period_unit_of_measure_values + r"])(\d*)$"
        )
        _feature_periods = []
        for fp in value:
            if isinstance(fp, FeaturePeriod):
                _feature_periods.append(fp)
            else:
                matches = re.match(regex, fp)
                if not matches:
                    raise FeaturePeriodMnemonicIsInvalid
                _feature_periods.append(
                    FeaturePeriod(
                        PeriodUnitOfMeasure(matches[2]),
                        int(matches[1]),
                        int(matches[3]),
                    )
                )
        self.__feature_periods = _feature_periods

    @property
    def features(self) -> list[Column]:
        # Find feature stems that do not correspond to any class in FEATURE_CLASSES.
        # If any are not found, raise an Exception.
        missing_stems = self.feature_stems - {
            cls.__name__ for cls in self.FEATURE_CLASSES
        }
        if missing_stems:
            # Only raise on the first (sorted for determinism)
            raise Exception(f"Feature(s) {sorted(missing_stems)} not found")
        desired_features = (
            [fc for fc in self.FEATURE_CLASSES if fc.__name__ in self.feature_stems]
            if self.feature_stems
            else self.FEATURE_CLASSES
        )
        return [
            feature.column
            for feature in [
                f[0](
                    as_at=self.as_at,
                    feature_period=f[1],
                    first_day_of_week=self.first_day_of_week,
                    use_absolute_periods=self.use_absolute_periods,
                )
                for f in (
                    (cls, fp) for cls in desired_features for fp in self.feature_periods
                )
            ]
        ]

    @property
    def references(self) -> dict[str, list[str]]:
        # this function requires a SparkSession in order to do its thing.
        # In normal operation a SparkSession will probably already exist
        # but in unit tests that might not be the case, so getOrCreate one
        SparkSession.builder.getOrCreate()
        return {
            # pylint: disable=protected-access
            node.name().head(): self.parse_references(node.toString())
            for node in [c._jc.node() for c in self.features]  # type: ignore
        }

    @property
    def flattened_references(self) -> set[str]:
        return {item for sublist in self.references.values() for item in sublist}

    @staticmethod
    def parse_references(references: str) -> list[str]:
        return sorted(
            set(re.findall(r"UnresolvedAttribute\(List\(([^)]+)\)", references))
        )

    def with_feature_periods(self, feature_periods: FeaturePeriodsType) -> Self:
        self.feature_periods = feature_periods
        return self

    def with_feature_period(self, feature_period: FeaturePeriod | str) -> Self:
        self.feature_periods = [*self.feature_periods, feature_period]
        return self

    def with_feature_stems(self, feature_stems: list[str]) -> Self:
        self.feature_stems = set[str](feature_stems)
        return self

    def with_feature_stem(self, feature_stem: str) -> Self:
        self.feature_stems.add(feature_stem)
        return self

    def with_first_day_of_week(self, first_day_of_week: str) -> Self:
        self.first_day_of_week = first_day_of_week
        return self

    def with_as_at(self, as_at: date) -> Self:
        self.as_at = as_at
        return self

    def with_use_absolute_periods(self, use_absolute_periods: bool) -> Self:
        self.use_absolute_periods = use_absolute_periods
        return self

    def without_feature_period(self, feature_period: FeaturePeriod | str) -> Self:
        if isinstance(feature_period, str):
            # Reuse setter to parse the mnemonic, then extract the result
            original = self.feature_periods
            self.feature_periods = [feature_period]
            feature_period = self.feature_periods[0]
            self.feature_periods = original
        self.feature_periods = [
            fp for fp in self.feature_periods if fp != feature_period
        ]
        return self

    def without_feature_stem(self, feature_stem: str) -> Self:
        self.feature_stems.discard(feature_stem)
        return self

    def __repr__(self) -> str:
        periods = sorted([fp.mnemonic for fp in self.feature_periods])
        stems = sorted(self.feature_stems) if self.feature_stems is not None else None
        return (
            f"{self.__class__.__name__}"
            f"(as_at={self.as_at}"
            f", feature_periods={periods}"
            f", feature_stems={stems}"
            f", first_day_of_week={self.first_day_of_week})"
        )



import uuid
from decimal import Decimal
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta


import pytest
from pyspark.sql import DataFrame, SparkSession, Row
import pyspark.sql.functions as f
from pyspark.sql.types import StructType
from jstark.grocery import GroceryFeatures
from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure
from jstark.sample.transactions import FakeGroceryTransactions
from jstark.sample.mealkit_orders import FakeMealkitOrders


@pytest.fixture(scope="session")
def spark_session() -> SparkSession:
    """Need a sparksession for most of the tests, so create it once"""
    return (
        SparkSession.builder.config("spark.driver.host", "127.0.0.1")
        .config("spark.driver.bindAddress", "127.0.0.1")
        # Default is 200 partitions which adds unnecessary overhead for small test data
        .config("spark.sql.shuffle.partitions", "1")
        .getOrCreate()
    )


@pytest.fixture(scope="session")
def as_at_timestamp() -> datetime:
    """return a datetime to be used as as_at. This can be used in other fixtures to
    generate datetimes relative to this datetime
    """
    # This is a Wednesday. That's important to know because
    # it affects week calculations.
    return datetime(2022, 11, 30, 10, 12, 13)


@pytest.fixture(scope="session")
def mealkit_orders_schema() -> StructType:
    return FakeMealkitOrders().mealkit_orders_schema


@pytest.fixture(scope="session")
def dataframe_of_faker_mealkit_orders() -> DataFrame:
    return FakeMealkitOrders(seed=42, number_of_orders=10000).df


@pytest.fixture(scope="session")
def purchases_schema() -> StructType:
    return FakeGroceryTransactions().transactions_schema


@pytest.fixture(scope="session")
def dataframe_of_faker_purchases() -> DataFrame:
    return FakeGroceryTransactions(seed=42, number_of_baskets=10000).df


@pytest.fixture(scope="session")
def dataframe_of_purchases(
    spark_session: SparkSession, as_at_timestamp: datetime, purchases_schema: StructType
) -> DataFrame:
    # We need lots of transactions of the same thing just with different days so
    # let's create the bulk of it once and then add the different days to it.
    chew_car_timestamp = as_at_timestamp - timedelta(days=2)
    chew_car = {
        "Timestamp": chew_car_timestamp,
        "Customer": "Chewie",
        "Store": "Hammersmith",
        "Channel": "Instore",
        "Basket": uuid.uuid4(),
        "items": [
            {
                "Product": "Carrot",
                "Quantity": 1,
                "GrossSpend": Decimal(0.1),
                "NetSpend": Decimal(0.1),
                "Discount": Decimal(0.0),
            },
        ],
    }
    transactions = [
        {
            "Timestamp": as_at_timestamp - relativedelta(months=12),
            "Customer": "Luke",
            "Store": "Ealing",
            "Channel": "Instore",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "Kleenex",
                    "Quantity": 1,
                    "GrossSpend": Decimal(8),
                    "NetSpend": Decimal(7.75),
                    "Discount": Decimal(0.5),
                }
            ],
        },
        {
            "Timestamp": as_at_timestamp - relativedelta(months=2),
            "Customer": "Luke",
            "Store": "Twickenham",
            "Channel": "Online",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "WD40",
                    "Quantity": 2,
                    "GrossSpend": Decimal(7),
                    "NetSpend": Decimal(6.75),
                    "Discount": Decimal(0.0),
                }
            ],
        },
        {
            "Timestamp": as_at_timestamp - relativedelta(months=1),
            "Customer": "Leia",
            "Store": "Ealing",
            "Channel": "Instore",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "Tiger Bread",
                    "Quantity": 3,
                    "GrossSpend": Decimal(6),
                    "NetSpend": Decimal(5.75),
                    "Discount": Decimal(0.1),
                }
            ],
        },
        {
            "Timestamp": as_at_timestamp - timedelta(days=4),
            "Customer": "Luke",
            "Store": "Ealing",
            "Channel": "Instore",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "Apples",
                    "Quantity": 6,
                    "GrossSpend": Decimal(3.25),
                    "NetSpend": Decimal(3.0),
                    "Discount": Decimal(0.75),
                }
            ],
        },
        {
            "Timestamp": as_at_timestamp - timedelta(days=1),
            "Customer": "Luke",
            "Store": "Ealing",
            "Channel": "Instore",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "Custard Creams",
                    "Quantity": 1,
                    "GrossSpend": Decimal(4.00),
                    "NetSpend": Decimal(3.75),
                    "Discount": Decimal(0.0),
                }
            ],
        },
        {
            "Timestamp": as_at_timestamp,
            "Customer": "Leia",
            "Store": "Hammersmith",
            "Channel": "Instore",
            "Basket": uuid.uuid4(),
            "items": [
                {
                    "Product": "Cheddar",
                    "Quantity": 2,
                    "GrossSpend": Decimal(2.50),
                    "NetSpend": Decimal(2.25),
                    "Discount": Decimal(0.1),
                },
                {
                    "Product": "Grapes",
                    "Quantity": 1,
                    "GrossSpend": Decimal(3.00),
                    "NetSpend": Decimal(2.75),
                    "Discount": Decimal(0.1),
                },
            ],
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=3),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=5),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=6),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=9),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=10),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=12),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=13),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=15),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=16),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=17),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=18),
        },
        {
            **chew_car,
            "Basket": uuid.uuid4(),
            "Timestamp": chew_car_timestamp - relativedelta(months=21),
        },
    ]
    flattened_transactions = FakeGroceryTransactions.flatten_transactions(transactions)
    return spark_session.createDataFrame(
        flattened_transactions,
        schema=purchases_schema,  # type: ignore
    )


@pytest.fixture(scope="session")
def luke_and_leia_purchases(dataframe_of_purchases: DataFrame) -> DataFrame:
    return dataframe_of_purchases.where(
        (f.col("Customer") == "Leia") | (f.col("Customer") == "Luke")
    )


@pytest.fixture(scope="session")
def purchasing_feature_generator(
    as_at_timestamp: datetime,
) -> GroceryFeatures:
    return GroceryFeatures(
        as_at=as_at_timestamp.date(),
        feature_periods=[
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 0),
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 2),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 1),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 0),
            FeaturePeriod(PeriodUnitOfMeasure.MONTH, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 1),
            FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 0),
            FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 1, 1),
            FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 1, 0),
            FeaturePeriod(PeriodUnitOfMeasure.YEAR, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.YEAR, 1, 1),
            FeaturePeriod(PeriodUnitOfMeasure.YEAR, 1, 0),
            FeaturePeriod(PeriodUnitOfMeasure.YEAR, 2, 0),
        ],
    )


@pytest.fixture(scope="session")
def luke_and_leia_purchases_first(
    luke_and_leia_purchases: DataFrame,
    purchasing_feature_generator: GroceryFeatures,
) -> Row:
    """
    If we only collect once, the tests should run quicker
    """
    df = luke_and_leia_purchases.groupBy().agg(*purchasing_feature_generator.features)
    first = df.first()
    assert first is not None
    return first



from jstark.features.gross_spend import GrossSpend
from jstark.features.net_spend import NetSpend
from jstark.features.count import Count
from jstark.features.recency_days import RecencyDays
from .basket_count import BasketCount
from .store_count import StoreCount
from jstark.features.customer_count import CustomerCount
from .channel_count import ChannelCount
from .approx_basket_count import ApproxBasketCount
from jstark.features.approx_customer_count import ApproxCustomerCount
from jstark.features.min_gross_spend import MinGrossSpend
from jstark.features.max_gross_spend import MaxGrossSpend
from jstark.features.min_net_spend import MinNetSpend
from jstark.features.max_net_spend import MaxNetSpend
from .average_gross_spend_per_basket import AvgGrossSpendPerBasket
from jstark.features.quantity import Quantity
from .average_quantity_per_basket import AvgQuantityPerBasket
from jstark.features.most_recent_purchase_date import MostRecentPurchaseDate
from jstark.features.min_net_price import MinNetPrice
from jstark.features.max_net_price import MaxNetPrice
from jstark.features.min_gross_price import MinGrossPrice
from jstark.features.max_gross_price import MaxGrossPrice
from jstark.features.earliest_purchase_date import EarliestPurchaseDate
from .average_discount_per_basket import AvgDiscountPerBasket
from .average_purchase_cycle import AvgPurchaseCycle
from .cycles_since_last_purchase import CyclesSinceLastPurchase
from .basket_periods import BasketPeriods
from .recency_weighted_basket import (
    RecencyWeightedBasket90,
    RecencyWeightedBasket95,
    RecencyWeightedBasket99,
    RecencyWeightedApproxBasket90,
    RecencyWeightedApproxBasket95,
    RecencyWeightedApproxBasket99,
)
from .average_basket import AvgBasket
from .grocery_features import GroceryFeatures

__all__ = [
    "GrossSpend",
    "NetSpend",
    "Count",
    "RecencyDays",
    "BasketCount",
    "StoreCount",
    "ProductCount",
    "CustomerCount",
    "ChannelCount",
    "ApproxBasketCount",
    "ApproxCustomerCount",
    "Discount",
    "MinGrossSpend",
    "MaxGrossSpend",
    "MinNetSpend",
    "MaxNetSpend",
    "AvgGrossSpendPerBasket",
    "Quantity",
    "AvgQuantityPerBasket",
    "MostRecentPurchaseDate",
    "MinNetPrice",
    "MaxNetPrice",
    "MinGrossPrice",
    "MaxGrossPrice",
    "EarliestPurchaseDate",
    "AvgDiscountPerBasket",
    "AvgPurchaseCycle",
    "CyclesSinceLastPurchase",
    "BasketPeriods",
    "RecencyWeightedBasket90",
    "RecencyWeightedBasket95",
    "RecencyWeightedBasket99",
    "RecencyWeightedApproxBasket90",
    "RecencyWeightedApproxBasket95",
    "RecencyWeightedApproxBasket99",
    "AvgBasket",
    "GroceryFeatures",
]



"""RecencyDays feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.min_feature import Min


class RecencyDays(Min):
    def column_expression(self) -> Column:
        return f.datediff(f.lit(self.as_at), f.col("Timestamp"))

    def default_value(self) -> Column:
        return f.lit(0)

    @property
    def description_subject(self) -> str:
        return "Minimum number of days since occurrence"

    @property
    def commentary(self) -> str:
        return (
            "This could be particularly useful (for example) in a grocery retailer "
            + "for determining when a customer most recently bought a product or "
            + " when a product was most recently bought in a store"
            + "Also note that this is very similar to "
            + f"MostRecentPurchaseDate_{self.feature_period.mnemonic} "
            + "so consider which of these "
            + "features is most useful to you."
        )



"""MinNetPrice feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.min_feature import Min


class MinNetPrice(Min):
    def column_expression(self) -> Column:
        return f.try_divide(f.col("NetSpend"), f.col("Quantity"))

    @property
    def description_subject(self) -> str:
        return "Minimum of (NetSpend / Quantity)"

    @property
    def commentary(self) -> str:
        return (
            "The net price is calculated as the net "
            + "spend dividied by how many were bought."
        )



"""AvgDiscountPerBasket feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.feature import DerivedFeature
from jstark.features.discount import Discount
from .basket_count import BasketCount


class AvgDiscountPerBasket(DerivedFeature):
    def column_expression(self) -> Column:
        return f.try_divide(
            Discount(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
            BasketCount(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
        )

    @property
    def description_subject(self) -> str:
        return "Average Discount per Basket"

    @property
    def commentary(self) -> str:
        return "Total Discount divided by the number of baskets."

    def default_value(self) -> Column:
        return f.lit(None)



"""NetSpend feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.sum_feature import Sum


class NetSpend(Sum):
    def column_expression(self) -> Column:
        return f.col("NetSpend")

    @property
    def description_subject(self) -> str:
        return "Sum of NetSpend"

    @property
    def commentary(self) -> str:
        return (
            "The definition of NetSpend can be whatever you want "
            + "it to be though typically its the price exclusive of any tax paid"
        )



"""Sum abstract base class"""

from abc import ABCMeta
from typing import Callable

import pyspark.sql.functions as f
from pyspark.sql import Column


from .feature import BaseFeature


class Sum(BaseFeature, metaclass=ABCMeta):
    def aggregator(self) -> Callable[[Column], Column]:
        return self.sum_aggregator

    def default_value(self) -> Column:
        return f.lit(0.0)



from datetime import date

from jstark.feature_generator import FeaturePeriodsType
from jstark.features.count import Count
from jstark.features.net_spend import NetSpend
from jstark.features.gross_spend import GrossSpend
from jstark.features.recency_days import RecencyDays
from jstark.grocery.basket_count import BasketCount
from jstark.grocery.store_count import StoreCount
from jstark.features.product_count import ProductCount
from jstark.features.approx_product_count import ApproxProductCount
from jstark.features.customer_count import CustomerCount
from jstark.grocery.channel_count import ChannelCount
from jstark.features.approx_customer_count import ApproxCustomerCount
from jstark.grocery.approx_basket_count import ApproxBasketCount
from jstark.features.discount import Discount
from jstark.features.min_gross_spend import MinGrossSpend
from jstark.features.max_gross_spend import MaxGrossSpend
from jstark.features.min_net_spend import MinNetSpend
from jstark.features.max_net_spend import MaxNetSpend
from jstark.grocery.average_gross_spend_per_basket import AvgGrossSpendPerBasket
from jstark.features.quantity import Quantity
from jstark.grocery.average_quantity_per_basket import AvgQuantityPerBasket
from jstark.features.most_recent_purchase_date import MostRecentPurchaseDate
from jstark.features.min_net_price import MinNetPrice
from jstark.features.max_net_price import MaxNetPrice
from jstark.features.min_gross_price import MinGrossPrice
from jstark.features.max_gross_price import MaxGrossPrice
from jstark.features.earliest_purchase_date import EarliestPurchaseDate
from jstark.grocery.average_discount_per_basket import AvgDiscountPerBasket
from jstark.grocery.average_purchase_cycle import AvgPurchaseCycle
from jstark.grocery.cycles_since_last_purchase import CyclesSinceLastPurchase
from jstark.grocery.basket_periods import BasketPeriods
from jstark.grocery.recency_weighted_basket import (
    RecencyWeightedBasket99,
    RecencyWeightedBasket90,
    RecencyWeightedBasket95,
    RecencyWeightedApproxBasket90,
    RecencyWeightedApproxBasket95,
    RecencyWeightedApproxBasket99,
)
from jstark.grocery.average_basket import AvgBasket
from jstark.features.feature import Feature
from jstark.feature_generator import FeatureGenerator


class GroceryFeatures(FeatureGenerator):
    def __init__(
        self,
        as_at: date,
        feature_periods: FeaturePeriodsType = None,
        feature_stems: set[str] | list[str] | None = None,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at,
            feature_periods,
            feature_stems,
            first_day_of_week,
            use_absolute_periods,
        )

    FEATURE_CLASSES: set[type[Feature]] = {
        Count,
        NetSpend,
        GrossSpend,
        RecencyDays,
        BasketCount,
        StoreCount,
        ProductCount,
        CustomerCount,
        ChannelCount,
        ApproxBasketCount,
        ApproxProductCount,
        ApproxCustomerCount,
        Discount,
        MinGrossSpend,
        MaxGrossSpend,
        MinNetSpend,
        MaxNetSpend,
        AvgGrossSpendPerBasket,
        Quantity,
        AvgQuantityPerBasket,
        MostRecentPurchaseDate,
        MinNetPrice,
        MaxNetPrice,
        MinGrossPrice,
        MaxGrossPrice,
        EarliestPurchaseDate,
        AvgDiscountPerBasket,
        AvgPurchaseCycle,
        CyclesSinceLastPurchase,
        BasketPeriods,
        RecencyWeightedBasket95,
        RecencyWeightedBasket90,
        RecencyWeightedBasket99,
        RecencyWeightedApproxBasket95,
        RecencyWeightedApproxBasket90,
        RecencyWeightedApproxBasket99,
        AvgBasket,
    }



"""ApproxDistinctCount feature"""

from abc import ABCMeta
from typing import Callable

import pyspark.sql.functions as f
from pyspark.sql import Column

from .feature import BaseFeature


class ApproxDistinctCount(BaseFeature, metaclass=ABCMeta):
    def aggregator(self) -> Callable[[Column], Column]:
        return self.approx_count_distinct_aggregator

    def default_value(self) -> Column:
        return f.lit(0)



import random
from functools import cached_property
import uuid
from datetime import date
from typing import Any, Iterable
from decimal import Decimal
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
    DecimalType,
)
from faker import Faker
from faker.providers import DynamicProvider


class FakeMealkitOrders:
    def __init__(self, seed: int | None = None, number_of_orders: int = 1000):
        self.seed = seed
        self.number_of_orders = number_of_orders

    @property
    def mealkit_orders_schema(self) -> StructType:
        return StructType(
            [
                StructField("Timestamp", TimestampType(), True),
                StructField("Customer", StringType(), True),
                StructField("Product", StringType(), True),
                StructField("Recipe", StringType(), True),
                StructField("Cuisine", StringType(), True),
                StructField("Quantity", IntegerType(), True),
                StructField("Order", StringType(), True),
                StructField("Discount", DecimalType(10, 2), True),
            ]
        )

    @staticmethod
    def flatten_mealkit_orders(mealkit_orders: list[Any]) -> Iterable[dict[str, Any]]:
        return [
            {
                "Customer": d["Customer"],
                "Product": d["Product"],
                "Cuisine": d["Cuisine"],
                "Order": d["Order"],
                "Timestamp": d["Timestamp"],
                **d2,
            }
            for d in mealkit_orders
            for d2 in d["Recipes"]
        ]

    @cached_property
    def df(self) -> DataFrame:

        products_provider = DynamicProvider(
            provider_name="product",
            elements=[
                "classic-plan",
                "preset-box-bc",
                "balanced-living-t1",
                "dinner-box",
                "classic-plan-t11",
                "classic-plan-t12",
                "classic-plan-t13",
                "classic-plan-t14",
                "classic-plan-t15",
            ],
        )
        recipes_provider = DynamicProvider(
            provider_name="recipe",
            elements=[
                "Banging bangers and mash",
                "Fish and chips",
                "Pizza and salad",
                "Chicken curry",
                "Beef stew",
                "Vegetable lasagna",
                "Salad and bread",
                "Soup and bread",
                "Pasta and sauce",
            ],
        )
        cuisines_provider = DynamicProvider(
            provider_name="cuisine",
            elements=[
                "Italian",
                "French",
                "Spanish",
            ],
        )

        fake = Faker()
        if self.seed:
            Faker.seed(self.seed)

        products_fake = Faker()
        products_fake.add_provider(products_provider)

        recipes_fake = Faker()
        recipes_fake.add_provider(recipes_provider)

        cuisines_fake = Faker()
        cuisines_fake.add_provider(cuisines_provider)

        mealkit_orders = []

        possible_quantities = [1, 2]
        if self.seed:
            random.seed(self.seed)
        quantities = random.choices(
            possible_quantities,
            weights=[100, 1],
            k=self.number_of_orders * len(recipes_provider.elements),
        )
        for order in range(self.number_of_orders):
            recipes = []
            # if self.seed:
            #     random.seed(self.seed)
            for recipe_index in range(random.randint(2, 5)):
                r = recipes_fake.unique.recipe()
                quantity = quantities[(order * len(possible_quantities)) + recipe_index]
                recipes.append(
                    {
                        "Recipe": r,
                        "Quantity": quantity,
                    }
                )
            mealkit_orders.append(
                {
                    "Customer": fake.name(),
                    "Timestamp": fake.date_time_between(
                        start_date=date(2021, 1, 1), end_date=date(2021, 12, 31)
                    ),
                    "Order": str(uuid.uuid4()),
                    "Product": products_fake.product(),
                    "Cuisine": cuisines_fake.cuisine(),
                    "Recipes": recipes,
                    "Discount": Decimal(random.uniform(0, 5)),
                }
            )
            recipes_fake.unique.clear()
            cuisines_fake.unique.clear()
        flattened_mealkit_orders = self.flatten_mealkit_orders(mealkit_orders)
        spark = SparkSession.builder.getOrCreate()
        return spark.createDataFrame(
            flattened_mealkit_orders,
            schema=self.mealkit_orders_schema,  # type: ignore
        )



from .feature import BaseFeature
from .count import Count
from .customer_count import CustomerCount
from .approx_customer_count import ApproxCustomerCount
from .earliest_purchase_date import EarliestPurchaseDate
from .most_recent_purchase_date import MostRecentPurchaseDate
from .quantity import Quantity
from .recency_days import RecencyDays
from .gross_spend import GrossSpend
from .max_gross_price import MaxGrossPrice
from .min_gross_price import MinGrossPrice
from .max_gross_spend import MaxGrossSpend
from .min_gross_spend import MinGrossSpend
from .net_spend import NetSpend
from .max_net_price import MaxNetPrice
from .max_net_spend import MaxNetSpend
from .min_net_price import MinNetPrice
from .min_net_spend import MinNetSpend

__all__ = [
    "BaseFeature",
    "Count",
    "CustomerCount",
    "ApproxCustomerCount",
    "EarliestPurchaseDate",
    "MostRecentPurchaseDate",
    "Quantity",
    "RecencyDays",
    "GrossSpend",
    "MaxGrossPrice",
    "MinGrossPrice",
    "MaxGrossSpend",
    "MinGrossSpend",
    "NetSpend",
    "MaxNetPrice",
    "MaxNetSpend",
    "MinNetPrice",
    "MinNetSpend",
]



"""MaxGrossSpend feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.max_feature import Max


class MaxGrossSpend(Max):
    def column_expression(self) -> Column:
        return f.col("GrossSpend")

    @property
    def description_subject(self) -> str:
        return "Maximum GrossSpend value"



"""Feature abstract base class

All feature classes are derived from here
"""

from abc import ABCMeta, abstractmethod
from datetime import date, timedelta, datetime
from typing import Callable
from dateutil.relativedelta import relativedelta


from pyspark.sql import Column
import pyspark.sql.functions as f
import pendulum

from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure
from jstark.features.first_and_last_date_of_period import FirstAndLastDateOfPeriod
from jstark.exceptions import AsAtIsNotADate


class Feature(metaclass=ABCMeta):
    def __init__(
        self,
        as_at: date,
        feature_period: FeaturePeriod,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        self.feature_period = feature_period
        if isinstance(as_at, datetime):
            import warnings

            warnings.warn(f"as_at={as_at!r} was converted to a date")
            as_at = as_at.date()
        if not isinstance(as_at, date):
            raise AsAtIsNotADate
        self._as_at = as_at
        self._first_day_of_week = first_day_of_week
        self._use_absolute_periods = use_absolute_periods

    @property
    def feature_period(self) -> FeaturePeriod:
        return self._feature_period

    @feature_period.setter
    def feature_period(self, value) -> None:
        self._feature_period = value

    @property
    def as_at(self) -> date:
        return self._as_at

    @property
    def feature_name(self) -> str:
        suffix = (
            self.column_metadata["period-absolute"]
            if self._use_absolute_periods
            else self.feature_period.mnemonic
        )
        return f"{type(self).__name__}_{suffix}"

    @property
    @abstractmethod
    def column(self) -> Column:
        """Complete definition of the column returned by this feature,
        replete with feature period filtering, metadata, default value
        and alias"""

    @property
    @abstractmethod
    def description_subject(self) -> str:
        """Desciption of the feature that will be concatenated
        with an explanation of the feature period.
        """

    @property
    def commentary(self) -> str:
        return "No commentary supplied"

    @abstractmethod
    def default_value(self) -> Column:
        """Default value of the feature, typically used when zero rows match
        the feature's feature_period
        """

    @abstractmethod
    def column_expression(self) -> Column:
        """The expression that defines the feature"""

    @property
    def start_date(self) -> date:
        n_days_ago = self.as_at - timedelta(days=self.feature_period.start)
        n_weeks_ago = self.as_at - timedelta(weeks=self.feature_period.start)
        n_months_ago = self.as_at - relativedelta(months=self.feature_period.start)
        n_quarters_ago = self.as_at - relativedelta(
            months=self.feature_period.start * 3
        )
        n_years_ago = self.as_at - relativedelta(years=self.feature_period.start)
        match self.feature_period.period_unit_of_measure:
            case PeriodUnitOfMeasure.DAY:
                return n_days_ago
            case PeriodUnitOfMeasure.WEEK:
                return FirstAndLastDateOfPeriod(
                    n_weeks_ago, self._first_day_of_week
                ).first_date_in_week
            case PeriodUnitOfMeasure.MONTH:
                return FirstAndLastDateOfPeriod(n_months_ago).first_date_in_month
            case PeriodUnitOfMeasure.QUARTER:
                return FirstAndLastDateOfPeriod(n_quarters_ago).first_date_in_quarter
            case _:  # PeriodUnitOfMeasure.YEAR:
                return FirstAndLastDateOfPeriod(n_years_ago).first_date_in_year

    @property
    def end_date(self) -> date:
        n_days_ago = self.as_at - timedelta(days=self.feature_period.end)
        n_weeks_ago = self.as_at - timedelta(weeks=self.feature_period.end)
        n_months_ago = self.as_at - relativedelta(months=self.feature_period.end)
        n_quarters_ago = self.as_at - relativedelta(months=self.feature_period.end * 3)
        n_years_ago = self.as_at - relativedelta(years=self.feature_period.end)
        match self.feature_period.period_unit_of_measure:
            case PeriodUnitOfMeasure.DAY:
                last_day_of_period = n_days_ago
            case PeriodUnitOfMeasure.WEEK:
                last_day_of_period = FirstAndLastDateOfPeriod(
                    n_weeks_ago, self._first_day_of_week
                ).last_date_in_week
            case PeriodUnitOfMeasure.MONTH:
                last_day_of_period = FirstAndLastDateOfPeriod(
                    n_months_ago
                ).last_date_in_month
            case PeriodUnitOfMeasure.QUARTER:
                last_day_of_period = FirstAndLastDateOfPeriod(
                    n_quarters_ago
                ).last_date_in_quarter
            case _:  # PeriodUnitOfMeasure.YEAR:
                last_day_of_period = FirstAndLastDateOfPeriod(
                    n_years_ago
                ).last_date_in_year
        # min() is used to ensure we don't return a date later than self.as_at
        return min(last_day_of_period, self.as_at)

    @property
    def column_metadata(self) -> dict[str, str]:
        period_absolute_start_period: str = ""
        period_absolute_end_period: str = ""
        match self.feature_period.period_unit_of_measure:
            case PeriodUnitOfMeasure.DAY:
                period_absolute_start_period = (
                    pendulum.instance(self.as_at)
                    .subtract(days=self.feature_period.start)
                    .format("YYYYMMDD")
                )
                period_absolute_end_period = (
                    pendulum.instance(self.as_at)
                    .subtract(days=self.feature_period.end)
                    .format("YYYYMMDD")
                )
            case PeriodUnitOfMeasure.WEEK:
                period_absolute_start_period = self._week_label(self.start_date)
                end_week_start = FirstAndLastDateOfPeriod(
                    pendulum.instance(self.as_at).subtract(
                        weeks=self.feature_period.end
                    ),
                    self._first_day_of_week,
                ).first_date_in_week
                period_absolute_end_period = self._week_label(end_week_start)
            case PeriodUnitOfMeasure.MONTH:
                period_absolute_start_period = (
                    pendulum.instance(self.as_at)
                    .subtract(months=self.feature_period.start)
                    .format("YYYYMMM")
                )
                period_absolute_end_period = (
                    pendulum.instance(self.as_at)
                    .subtract(months=self.feature_period.end)
                    .format("YYYYMMM")
                )
            case PeriodUnitOfMeasure.QUARTER:
                dt_start = pendulum.instance(self.as_at).subtract(
                    months=self.feature_period.start * 3
                )
                period_absolute_start_period = f"{dt_start.year}Q{dt_start.quarter}"
                dt_end = pendulum.instance(self.as_at).subtract(
                    months=self.feature_period.end * 3
                )
                period_absolute_end_period = f"{dt_end.year}Q{dt_end.quarter}"
            case _:  # PeriodUnitOfMeasure.YEAR:
                period_absolute_start_period = str(
                    pendulum.instance(self.as_at)
                    .subtract(years=self.feature_period.start)
                    .year
                )
                period_absolute_end_period = str(
                    pendulum.instance(self.as_at)
                    .subtract(years=self.feature_period.end)
                    .year
                )
        return {
            "created-with-love-by": "https://github.com/jamiekt/jstark",
            "start-date": self.start_date.strftime("%Y-%m-%d"),
            "end-date": self.end_date.strftime("%Y-%m-%d"),
            "description": (
                f"{self.description_subject} between "
                + f"{self.start_date.strftime('%Y-%m-%d')} and "
                + f"{self.end_date.strftime('%Y-%m-%d')}"
            ),
            "generated-at": datetime.now().strftime("%Y-%m-%d"),
            "commentary": self.commentary,
            "name-stem": str(type(self).__name__),
            "period-absolute-start": period_absolute_start_period,
            "period-absolute-end": period_absolute_end_period,
            "period-absolute": period_absolute_start_period
            if period_absolute_start_period == period_absolute_end_period
            else f"{period_absolute_start_period}-{period_absolute_end_period}",
        }

    def _week_label(self, week_start_date: date) -> str:
        """Convert the first day of a week to a label like 2026W13.

        W01 of a year starts on the first occurrence of first_day_of_week
        on or before Jan 1 of that year.
        """
        weekdays = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        first_day_of_week = self._first_day_of_week or "Monday"
        target_weekday = weekdays.index(first_day_of_week)
        year = week_start_date.year

        jan1 = date(year, 1, 1)
        days_back = (jan1.weekday() - target_weekday) % 7
        w01_start = pendulum.instance(jan1).subtract(days=days_back)

        jan1_next = date(year + 1, 1, 1)
        days_back_next = (jan1_next.weekday() - target_weekday) % 7
        w01_start_next = pendulum.instance(jan1_next).subtract(days=days_back_next)

        if week_start_date >= w01_start_next:
            w01_start = w01_start_next
            year = year + 1

        week_number = (week_start_date - w01_start).days // 7 + 1
        return f"{year}W{week_number:02d}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(as_at={self.as_at}"
            f", feature_period='{self.feature_period.mnemonic}'"
            f", first_day_of_week={self._first_day_of_week!r})"
        )


class DerivedFeature(Feature, metaclass=ABCMeta):
    """A DerivedFeature is a feature that is calculated by combining
    data that has already been aggregated. For example, a derived
    feature called 'Average Gross Spend Per Basket' would be calculated
    by dividing the total GrossSpend by number of baskets (BasketCount)
    """

    @property
    def column(self) -> Column:
        return f.coalesce(self.column_expression(), self.default_value()).alias(
            self.feature_name, metadata=self.column_metadata
        )


class BaseFeature(Feature, metaclass=ABCMeta):
    """A BaseFeature is a feature that is calculated by aggregating
    raw source data. That data may have been cleaned and transformed in
    some way, but typically the grain of that data is real occurrences
    of some activity. Examples of such data are lists of grocery
    transactions, phone calls or journeys.
    """

    def sum_aggregator(self, column: Column) -> Column:
        return f.sum(column)

    def count_aggregator(self, column: Column) -> Column:
        return f.count(column)

    def count_if_aggregator(self, column: Column) -> Column:
        return f.count_if(column)

    def count_distinct_aggregator(self, column: Column) -> Column:
        return f.countDistinct(column)

    def approx_count_distinct_aggregator(self, column: Column) -> Column:
        return f.approx_count_distinct(column)

    def max_aggregator(self, column: Column) -> Column:
        return f.max(column)

    def min_aggregator(self, column: Column) -> Column:
        return f.min(column)

    def collect_set_aggregator(self, column: Column) -> Column:
        return f.collect_set(column)

    @abstractmethod
    def aggregator(self) -> Callable[[Column], Column]:
        """Aggregator function"""

    @property
    def column(self) -> Column:
        return f.coalesce(
            self.aggregator()(
                f.when(
                    (f.to_date(f.col("Timestamp")) >= f.lit(self.start_date))
                    & (f.to_date(f.col("Timestamp")) <= f.lit(self.end_date)),
                    self.column_expression(),
                )
            ),
            self.default_value(),
        ).alias(self.feature_name, metadata=self.column_metadata)



"""OrderCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class OrderCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Order")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Orders"

    @property
    def commentary(self) -> str:
        return (
            "The number of orders. Typically the dataframe supplied "
            + "to this feature will have many recipes for the same order, "
            + "this feature allows you to determine how many distinct orders "
            + "existed in that collection of orders."
        )



from datetime import date
from jstark import feature_generator
from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure


def test_with_feature_periods():
    fg = feature_generator.FeatureGenerator().with_feature_periods(["3m1", "6m4"])
    assert fg.feature_periods == [
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 3, 1),
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 6, 4),
    ]


def test_with_feature_period():
    fg1 = feature_generator.FeatureGenerator().with_feature_period("10w10")
    fg2 = feature_generator.FeatureGenerator(feature_periods=["10w10"])
    expected = FeaturePeriod(PeriodUnitOfMeasure.WEEK, 10, 10)
    assert expected in fg1.feature_periods
    assert expected in fg2.feature_periods


def test_without_feature_period():
    fg = feature_generator.FeatureGenerator().with_feature_periods(["3m1", "6m4"])
    fg = fg.without_feature_period(FeaturePeriod(PeriodUnitOfMeasure.MONTH, 3, 1))
    assert FeaturePeriod(PeriodUnitOfMeasure.MONTH, 3, 1) not in fg.feature_periods


def test_without_feature_period_mnemonic():
    fg = feature_generator.FeatureGenerator().with_feature_periods(["3m1", "6m4"])
    fg = fg.without_feature_period("3m1")
    assert FeaturePeriod(PeriodUnitOfMeasure.MONTH, 3, 1) not in fg.feature_periods


def test_with_as_at():
    fg = feature_generator.FeatureGenerator()
    assert fg.as_at == date.today()
    fg = fg.with_as_at(date(2026, 3, 21))
    assert fg.as_at == date(2026, 3, 21)


def test_with_first_day_of_week():
    fg = feature_generator.FeatureGenerator()
    assert fg.first_day_of_week is None
    fg = fg.with_first_day_of_week("Monday")
    assert fg.first_day_of_week == "Monday"


def test_with_use_absolute_periods():
    fg = feature_generator.FeatureGenerator()
    assert fg.use_absolute_periods is False
    fg = fg.with_use_absolute_periods(True)
    assert fg.use_absolute_periods is True



"""CustomerCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class CustomerCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Customer")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Customers"

    @property
    def commentary(self) -> str:
        return (
            "The number of customers. Typically the dataframe supplied "
            + "to this feature will have many transactions for many baskets each bought"
            + " a different customer, this feature allows you to determine how many "
            + "disinct customerss bought those baskets."
        )



from datetime import date, datetime, timedelta
import pytest
from math import pow
from jstark.grocery.recency_weighted_basket import RecencyWeightedBasket
from pyspark.sql import DataFrame, Row
import pyspark.sql.functions as f

from jstark.grocery import GroceryFeatures, BasketCount
from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure
from jstark.exceptions import FeaturePeriodMnemonicIsInvalid


def test_feature_period_mnemonic():
    pfg1 = GroceryFeatures(date.today(), [FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0)])
    pfg2 = GroceryFeatures(date.today(), ["0d0"])
    assert (
        pfg1.feature_periods[0].period_unit_of_measure.value
        == pfg2.feature_periods[0].period_unit_of_measure.value
    )
    assert pfg1.feature_periods[0].start == pfg2.feature_periods[0].start
    assert pfg1.feature_periods[0].end == pfg2.feature_periods[0].end


def test_feature_period_invalid_mnemonic():
    with pytest.raises(FeaturePeriodMnemonicIsInvalid) as exc_info:
        GroceryFeatures(date.today(), ["qwerty"])
    assert str(exc_info.value) == (
        "FeaturePeriod mnemonic must be an integer followed by a letter "
        + "from ['d', 'w', 'm', 'q', 'y'] followed by an integer"
    )


def test_feature_period_invalid_mnemonic_empty_string():
    with pytest.raises(FeaturePeriodMnemonicIsInvalid) as exc_info:
        GroceryFeatures(date.today(), [""])
    assert str(exc_info.value) == (
        "FeaturePeriod mnemonic must be an integer followed by a letter "
        + "from ['d', 'w', 'm', 'q', 'y'] followed by an integer"
    )


def test_feature_period_invalid_mnemonic_unit_of_measure():
    with pytest.raises(FeaturePeriodMnemonicIsInvalid) as exc_info:
        GroceryFeatures(date.today(), ["0z0"])
    assert str(exc_info.value) == (
        "FeaturePeriod mnemonic must be an integer followed by a letter "
        + "from ['d', 'w', 'm', 'q', 'y'] followed by an integer"
    )


def test_parse_references_grossspend():
    """
    PurchasingFeatureGenerator.parse_references() is only an internal helper function
    but still a good idea to have a couple of little tests here to verify and
    document its behaviour
    """
    assert GroceryFeatures.parse_references(
        "Alias(UnresolvedAttribute(List(GrossSpend),None,false,Origin()),"
        "UnresolvedAttribute(List(Timestamp),None,false,Origin()))"
    ) == ["GrossSpend", "Timestamp"]


def test_parse_references_count():
    """Basic test that static method parse_references works"""
    assert GroceryFeatures.parse_references(
        "Alias(UnresolvedAttribute(List(Timestamp),None,false,Origin()))"
    ) == ["Timestamp"]


def test_references_count(
    purchasing_feature_generator: GroceryFeatures,
):
    """Verify that Count_0d0 requires field Timestamp"""
    assert purchasing_feature_generator.references["Count_0d0"] == ["Timestamp"]


def test_references_grossspend(
    purchasing_feature_generator: GroceryFeatures,
):
    """Verify that Count_0d0 requires fields Timestamp,GrossSpend"""
    assert purchasing_feature_generator.references["GrossSpend_0d0"] == [
        "GrossSpend",
        "Timestamp",
    ]


def test_count_today(luke_and_leia_purchases_first: Row):
    """Test Count_0d0"""
    assert float(luke_and_leia_purchases_first["Count_0d0"]) == 2


def test_flattened_references(purchasing_feature_generator: GroceryFeatures):
    assert purchasing_feature_generator.flattened_references == {
        "GrossSpend",
        "Timestamp",
        "Basket",
        "Channel",
        "Store",
        "NetSpend",
        "Discount",
        "Product",
        "Quantity",
        "Customer",
    }


def test_gross_spend_today(
    as_at_timestamp: datetime, luke_and_leia_purchases: DataFrame
):
    """Test GrossSpend_0d0"""

    def _validate_expected_value(df, expected):
        result = df.first()
        assert result is not None
        assert float(result["GrossSpend_0d0"]) == expected
        return result

    df = luke_and_leia_purchases.groupBy().agg(
        *GroceryFeatures(
            as_at=as_at_timestamp.date(),
            feature_periods=[
                FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
            ],
        ).features
    )
    _validate_expected_value(df, 5.5)
    df = luke_and_leia_purchases.groupBy().agg(
        *GroceryFeatures(
            as_at=as_at_timestamp.date() - timedelta(days=1),
            feature_periods=[
                FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
            ],
        ).features
    )
    _validate_expected_value(df, 4.0)


def test_gross_spend_today_and_yesterday(
    as_at_timestamp: datetime, luke_and_leia_purchases: DataFrame
):
    """Test GrossSpend_1d0"""

    def _validate_expected_value(df, expected):
        result = df.first()
        assert result is not None
        assert float(result["GrossSpend_1d0"]) == expected
        return result

    df = luke_and_leia_purchases.groupBy().agg(
        *GroceryFeatures(
            as_at=as_at_timestamp.date(),
            feature_periods=[
                FeaturePeriod(PeriodUnitOfMeasure.DAY, 1, 0),
            ],
        ).features
    )
    _validate_expected_value(df, 9.5)
    df = luke_and_leia_purchases.groupBy().agg(
        *GroceryFeatures(
            as_at=as_at_timestamp.date() - timedelta(days=1),
            feature_periods=[
                FeaturePeriod(PeriodUnitOfMeasure.DAY, 1, 0),
            ],
        ).features
    )
    _validate_expected_value(df, 4.0)


def test_gross_spend_today_yesterday_and_day_before_yesterday(
    luke_and_leia_purchases_first: Row,
):
    """This test is being used to verify the logic that filters the
    input dataframe according to each the respective feature period
    """
    assert float(luke_and_leia_purchases_first["GrossSpend_2d0"]) == 9.5


def test_gross_spend_day_before_yesterday(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_2d2"""
    assert float(luke_and_leia_purchases_first["GrossSpend_2d2"]) == 0


def test_gross_spend_this_week(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_0w0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_0w0"]) == 9.5


def test_gross_spend_last_week(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1w1"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1w1"]) == 3.25


def test_gross_spend_this_week_and_last_week(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1w0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1w0"]) == 12.75


def test_gross_spend_this_month(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_0m0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_0m0"]) == 12.75


def test_gross_spend_last_month(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1m1"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1m1"]) == 6.0


def test_gross_spend_this_month_and_last_month(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1m0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1m0"]) == 18.75


def test_gross_spend_this_quarter(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_0q0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_0q0"]) == 18.75


def test_gross_spend_last_quarter(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1q1"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1q1"]) == 7


def test_gross_spend_this_quarter_and_last_quarter(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1q0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1q0"]) == 25.75


def test_gross_spend_this_year(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_0y0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_0y0"]) == 25.75


def test_gross_spend_last_year(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1y1"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1y1"]) == 8


def test_gross_spend_this_year_and_last_year(luke_and_leia_purchases_first: Row):
    """Test GrossSpend_1y0"""
    assert float(luke_and_leia_purchases_first["GrossSpend_1y0"]) == 33.75


def test_grossspend_metadata_description(
    dataframe_of_purchases: DataFrame,
    purchasing_feature_generator: GroceryFeatures,
):
    """Test GrossSpend_2d0 metadata"""
    df = dataframe_of_purchases.agg(*purchasing_feature_generator.features)
    assert [c.metadata["description"] for c in df.schema if c.name == "GrossSpend_2d0"][
        0
    ] == "Sum of GrossSpend between 2022-11-28 and 2022-11-30"


def test_recencydays_and_mostrecentpurchasedate(
    dataframe_of_purchases: DataFrame,
    purchasing_feature_generator: GroceryFeatures,
):
    """Test RecencyDays_2y0 & MostRecentPurchaseDate_2y0"""
    dataframe_of_purchases = dataframe_of_purchases.where(
        f.col("Timestamp") < f.lit(date(2022, 1, 1))
    )
    df = dataframe_of_purchases.agg(*purchasing_feature_generator.features)
    first = df.select("RecencyDays_2y0", "MostRecentPurchaseDate_2y0").first()
    assert first is not None
    assert first["RecencyDays_2y0"] == 365
    assert first["MostRecentPurchaseDate_2y0"] == date(2021, 11, 30)


def test_basket_count_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test BasketCount_0y0"""
    assert float(luke_and_leia_purchases_first["BasketCount_0y0"]) == 5


def test_store_count_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test StoreCount_0y0"""
    assert float(luke_and_leia_purchases_first["StoreCount_0y0"]) == 3


def test_product_count_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test ProductCount_0y0"""
    assert float(luke_and_leia_purchases_first["ProductCount_0y0"]) == 6


def test_customer_count_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test CustomerCount_0y0"""
    assert float(luke_and_leia_purchases_first["CustomerCount_0y0"]) == 2


def test_channel_count_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test ChannelCount_0y0"""
    assert float(luke_and_leia_purchases_first["ChannelCount_0y0"]) == 2


def test_discount_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test Discount_0y0"""
    assert float(luke_and_leia_purchases_first["Discount_0y0"]) == 1.05


def test_net_spend_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test NetSpend_0y0"""
    assert float(luke_and_leia_purchases_first["NetSpend_0y0"]) == 24.25


def test_max_net_spend_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MaxNetSpend_0y0"""
    assert float(luke_and_leia_purchases_first["MaxNetSpend_0y0"]) == 6.75


def test_min_net_spend_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MinNetSpend_0y0"""
    assert float(luke_and_leia_purchases_first["MinNetSpend_0y0"]) == 2.25


def test_max_gross_spend_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MaxGrossSpend_0y0"""
    assert float(luke_and_leia_purchases_first["MaxGrossSpend_0y0"]) == 7.0


def test_min_gross_spend_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MinGrossSpend_0y0"""
    assert float(luke_and_leia_purchases_first["MinGrossSpend_0y0"]) == 2.5


def test_avg_gross_spend_per_basket_luke_and_leia_0y0(
    luke_and_leia_purchases_first: Row,
):
    """Test AvgGrossSpendPerBasket_0y0"""
    assert float(luke_and_leia_purchases_first["AvgGrossSpendPerBasket_0y0"]) == 5.15


def test_quantity_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test Quantity_0y0"""
    assert float(luke_and_leia_purchases_first["Quantity_0y0"]) == 15


def test_avg_quantity_per_basket_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test AvgQuantityPerBasket_0y0"""
    assert float(luke_and_leia_purchases_first["AvgQuantityPerBasket_0y0"]) == 3.0


def test_min_net_price_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MinNetPrice_0y0"""
    assert float(luke_and_leia_purchases_first["MinNetPrice_0y0"]) == 0.5


def test_max_net_price_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MaxNetPrice_0y0"""
    assert float(luke_and_leia_purchases_first["MaxNetPrice_0y0"]) == 3.75


def test_min_gross_price_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MinGrossPrice_0y0"""
    assert float(luke_and_leia_purchases_first["MinGrossPrice_0y0"]) == 0.5416666666667


def test_max_gross_price_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test MaxGrossPrice_0y0"""
    assert float(luke_and_leia_purchases_first["MaxGrossPrice_0y0"]) == 4.0


def test_earliest_purchase_date_luke_and_leia_0y0(luke_and_leia_purchases_first: Row):
    """Test EarliestPurchaseDate_2y0"""
    assert luke_and_leia_purchases_first["EarliestPurchaseDate_2y0"] == date(
        2021, 11, 30
    )


def test_basketweeks(
    as_at_timestamp: datetime, dataframe_of_faker_purchases: DataFrame
):
    """Test BasketWeeks

    BasketWeeks is the number of weeks in which at least one basket was purchased

    The result here is 5 because the input dataset includes data from 2021 whereas the
    specified as_at date is 2022-11-30, so there are only 5 weeks in the 52 weeks
    leading up to 2022-11-30 in which any baskets were purchased
    """
    pfg = GroceryFeatures(as_at=as_at_timestamp, feature_periods=["52w0"])
    output_df = (
        dataframe_of_faker_purchases.groupBy()
        .agg(*pfg.features)
        .select("BasketWeeks_52w0")
    )
    first = output_df.first()
    assert first is not None
    assert first["BasketWeeks_52w0"] == 5


def test_basketweeks_commentary(
    as_at_timestamp: datetime, dataframe_of_faker_purchases: DataFrame
):
    """Test BasketWeeks commentary"""
    pfg = GroceryFeatures(as_at=as_at_timestamp, feature_periods=["52w1"])
    output_df = (
        dataframe_of_faker_purchases.groupBy()
        .agg(*pfg.features)
        .select("BasketWeeks_52w1")
    )
    assert [(c.metadata["commentary"]) for c in output_df.schema][0] == (
        "The number of weeks in which at least one basket was purchased. "
        + "The value will be in the range 0 to 52"
        + " because 52 is the number of weeks between 2021-11-29 and 2022-11-27."
        + " When grouped by Customer and Product"
        + " this feature is a useful indicator of the frequency of which a"
        + " Customer purchases a Product."
    )


def get_dataframes_for_perweek_feature_tests(as_at_timestamp, luke_and_leia_purchases):
    fg = GroceryFeatures(
        as_at=as_at_timestamp.date(),
        feature_periods=[
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 13, 0),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 1),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 2, 2),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 3, 3),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 4, 4),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 5, 5),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 6, 6),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 7, 7),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 8, 8),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 9, 9),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 10, 10),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 11, 11),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 12, 12),
            FeaturePeriod(PeriodUnitOfMeasure.WEEK, 13, 13),
        ],
    )
    df = luke_and_leia_purchases.groupBy().agg(*fg.features)
    weeks_df = df.select(
        "BasketCount_0w0",
        "BasketCount_1w1",
        "BasketCount_2w2",
        "BasketCount_3w3",
        "BasketCount_4w4",
        "BasketCount_5w5",
        "BasketCount_6w6",
        "BasketCount_7w7",
        "BasketCount_8w8",
        "BasketCount_9w9",
        "BasketCount_10w10",
        "BasketCount_11w11",
        "BasketCount_12w12",
        "BasketCount_13w13",
    )
    df_first = df.first()
    weeks_df_first = weeks_df.first()
    assert df_first is not None
    assert weeks_df_first is not None
    return df_first, weeks_df_first


def test_recencyweightedbasketweeks_luke_and_leia(
    as_at_timestamp: datetime, luke_and_leia_purchases: DataFrame
):
    """Test RecencyWeightedBasketWeeks

    This test verifies the correct value of RecencyWeightedBasketWeeksXX by calculating
    the BasketCount for each individual week, calculating the smoothed value for that ,
    week then summing all those values. This is exactly the same calculation that is
    performed by the feature generator so it might be argued that this test doesn't add
    any value. I don't agree that that is the case however, it helps to demonstrate
    exactly what this feature provides and given that its not an easy one to explain, I
    think that has some value.

    A later addition to this test also verifies RecencyWeightedApproxBasketWeeksXX
    """
    df_first, df_weeks_first = get_dataframes_for_perweek_feature_tests(
        as_at_timestamp, luke_and_leia_purchases
    )
    recency_weighted_basket_count_weeks_90 = 0.0
    recency_weighted_basket_count_weeks_95 = 0.0
    recency_weighted_basket_count_weeks_99 = 0.0
    for i in range(14):
        # Loop over all the weeks, calculate the smoothed value for each,
        # then sum them all up
        recency_weighted_basket_count_weeks_90 += float(
            df_weeks_first[f"BasketCount_{i}w{i}"]
        ) * pow(0.9, i)
        recency_weighted_basket_count_weeks_95 += float(
            df_weeks_first[f"BasketCount_{i}w{i}"]
        ) * pow(0.95, i)
        recency_weighted_basket_count_weeks_99 += float(
            df_weeks_first[f"BasketCount_{i}w{i}"]
        ) * pow(0.99, i)
    assert recency_weighted_basket_count_weeks_90 == float(
        df_first["RecencyWeightedBasketWeeks90_13w0"]
    )
    assert recency_weighted_basket_count_weeks_95 == float(
        df_first["RecencyWeightedBasketWeeks95_13w0"]
    )
    assert recency_weighted_basket_count_weeks_99 == float(
        df_first["RecencyWeightedBasketWeeks99_13w0"]
    )
    # These features use approx_count_distinct (HyperLogLog) which is
    # non-deterministic, so the result can vary slightly across runs.
    # Use pytest.approx with abs=1.5 to allow for this.
    assert df_first["RecencyWeightedApproxBasketWeeks90_13w0"] == pytest.approx(
        3.943520489, abs=1.5
    )
    assert df_first["RecencyWeightedApproxBasketWeeks95_13w0"] == pytest.approx(
        4.394755659724609, abs=1.5
    )
    assert df_first["RecencyWeightedApproxBasketWeeks99_13w0"] == pytest.approx(
        4.864113257483641, abs=1.5
    )


def test_averagebasketsperweek_luke_and_leia(
    as_at_timestamp: datetime, luke_and_leia_purchases: DataFrame
):
    """Test AverageBasketsPerWeek"""
    df_first, df_weeks_first = get_dataframes_for_perweek_feature_tests(
        as_at_timestamp, luke_and_leia_purchases
    )
    n = 14
    total_baskets = sum(df_weeks_first[f"BasketCount_{i}w{i}"] for i in range(n))
    average_baskets_per_week = total_baskets / n
    assert average_baskets_per_week == df_first["AverageBasketsPerWeek_13w0"]


def test_averagepurchasecycle_chewie_carrot(
    dataframe_of_purchases: DataFrame,
    purchasing_feature_generator: GroceryFeatures,
):
    """
    Test AveragePurchaseCycle

    The input df has lots of transactions of chewie buying a carrot on different days.
    """
    input_df = dataframe_of_purchases.where(f.col("Customer") == "Chewie").where(
        f.col("Product") == "Carrot"
    )
    df = (
        input_df.groupBy(["Customer", "Product"])
        .agg(*purchasing_feature_generator.features)
        .select(
            [
                "Customer",
                "Product",
                "AvgPurchaseCycle_2y0",
                "CyclesSinceLastPurchase_2y0",
                # The following are not required for the assertions but they all
                # contribute to CyclesSinceLastPurchase so are useful for debugging.
                "MostRecentPurchaseDate_2y0",
                "EarliestPurchaseDate_2y0",
                "BasketCount_2y0",
                "RecencyDays_2y0",
            ]
        )
    )
    df_first = df.first()
    assert df_first is not None
    assert df_first["AvgPurchaseCycle_2y0"] == pytest.approx(45.5)
    assert df_first["CyclesSinceLastPurchase_2y0"] == pytest.approx(2.065934065934066)


def test_feature_generator_repr():
    gf = GroceryFeatures(
        as_at=date(2022, 11, 30),
        feature_periods=[
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 0),
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 2),
        ],
        feature_stems={"GrossSpend", "BasketCount"},
        first_day_of_week=None,
    )
    repr_string = repr(gf)
    # We can't control which order the feature stems appear in repr so we'll check
    # both possibilities
    feature_stems_str = "['BasketCount', 'GrossSpend']"
    repr_ = (
        "GroceryFeatures(as_at=2022-11-30, feature_periods=['0d0', '2d0', '2d2'], "
        + f"feature_stems={feature_stems_str}, first_day_of_week=None)"
    )
    assert repr_string == repr_


def test_feature_repr():
    bc = BasketCount(
        as_at=date(2022, 11, 30),
        feature_period=FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
        first_day_of_week=None,
    )
    repr_string = repr(bc)
    assert (
        repr_string
        == "BasketCount(as_at=2022-11-30, feature_period='0d0', first_day_of_week=None)"
    )

    rwb = RecencyWeightedBasket(
        as_at=date(2022, 11, 30),
        feature_period=FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
        smoothing_factor=0.9,
        first_day_of_week=None,
    )
    repr_string = repr(rwb)
    expected = (
        "RecencyWeightedBasket(as_at=2022-11-30"
        ", feature_period='0d0'"
        ", smoothing_factor=0.9"
        ", first_day_of_week=None)"
    )
    assert repr_string == expected


def test_use_absolute_periods(
    as_at_timestamp: datetime, luke_and_leia_purchases: DataFrame
):
    """Test that use_absolute_periods uses period-absolute as the feature name suffix"""
    fg = GroceryFeatures(
        as_at=as_at_timestamp.date(),
        feature_periods=[
            FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0),
        ],
        feature_stems=["Count"],
        use_absolute_periods=True,
    )
    df = luke_and_leia_purchases.groupBy().agg(*fg.features)
    assert "Count_20221130" in df.columns



"""Cuisines feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.collect_set_feature import CollectSet


class Cuisines(CollectSet):
    def column_expression(self) -> Column:
        return f.col("Cuisine")

    @property
    def description_subject(self) -> str:
        return "Set of Cuisines"

    @property
    def commentary(self) -> str:
        return (
            "The set of cuisines. Typically the dataframe supplied "
            + "to this feature will have many recipes for the same cuisine, "
            + "this feature allows you to determine the set of cuisines "
            + "that have been ordered."
        )



"AvgOrder feature"

from jstark.features.feature import DerivedFeature

from pyspark.sql import Column
import pyspark.sql.functions as f

from jstark.feature_period import FeaturePeriod
from .order_count import OrderCount


class AvgOrder(DerivedFeature):
    "Average orders per given periodunitofmeasure feature"

    def column_expression(self) -> Column:
        return (
            OrderCount(
                as_at=self.as_at,
                feature_period=FeaturePeriod(
                    self.feature_period.period_unit_of_measure,
                    self.feature_period.start,
                    self.feature_period.end,
                ),
                first_day_of_week=self._first_day_of_week,
            ).column
            / self.feature_period.number_of_periods
        )

    def default_value(self) -> Column:
        return f.lit(None)

    @property
    def description_subject(self) -> str:
        return (
            "Average number of orders per "
            + f"{self.feature_period.period_unit_of_measure.name.lower()}"
        )

    @property
    def feature_name(self) -> str:
        return (
            "AverageOrdersPer"
            + f"{self.feature_period.period_unit_of_measure.name.title()}"
            + f"_{self.feature_period.mnemonic}"
        )



from datetime import date

from jstark.feature_generator import FeaturePeriodsType
from jstark.features.count import Count
from jstark.features.recency_days import RecencyDays
from jstark.features.customer_count import CustomerCount
from jstark.features.approx_customer_count import ApproxCustomerCount
from jstark.features.quantity import Quantity
from jstark.features.product_count import ProductCount
from jstark.features.most_recent_purchase_date import MostRecentPurchaseDate
from jstark.features.earliest_purchase_date import EarliestPurchaseDate
from jstark.mealkit.order_periods import OrderPeriods
from jstark.mealkit.order_count import OrderCount
from jstark.mealkit.approx_order_count import ApproxOrderCount
from jstark.features.approx_product_count import ApproxProductCount
from jstark.mealkit.recipe_count import RecipeCount
from jstark.mealkit.approx_recipe_count import ApproxRecipeCount
from jstark.mealkit.average_order import AvgOrder
from jstark.features.discount import Discount
from jstark.features.feature import Feature
from jstark.feature_generator import FeatureGenerator
from jstark.mealkit.average_quantity_per_order import AvgQuantityPerOrder
from jstark.mealkit.cycles_since_last_order import CyclesSinceLastOrder
from jstark.mealkit.average_purchase_cycle import AvgPurchaseCycle
from jstark.mealkit.cuisines import Cuisines
from jstark.mealkit.cuisine_count import CuisineCount
from jstark.mealkit.cuisine_count import (
    ItalianCuisineCount,
    FrenchCuisineCount,
    SpanishCuisineCount,
    SrilankanCuisineCount,
    LebaneseCuisineCount,
    GreekCuisineCount,
    VietnameseCuisineCount,
    DanishCuisineCount,
    WesternEuropeCuisineCount,
    FusionCuisineCount,
    DutchCuisineCount,
    KoreanCuisineCount,
    SouthEastAsianCuisineCount,
    MalaysianCuisineCount,
    AmericanCuisineCount,
    AsianCuisineCount,
    IndianCuisineCount,
    GermanCuisineCount,
    CentralAmericaCuisineCount,
    ArgentinianCuisineCount,
    NorthAmericanCuisineCount,
    NewZealandCuisineCount,
    CanadianCuisineCount,
    SwedishCuisineCount,
    EgyptianCuisineCount,
    NorthernEuropeanCuisineCount,
    CajunCuisineCount,
    AustrianCuisineCount,
    MexicanCuisineCount,
    MediterraneanCuisineCount,
    ThaiCuisineCount,
    ChineseCuisineCount,
    LatinAmericanCuisineCount,
    SouthAsiaCuisineCount,
    MiddleEasternCuisineCount,
    TraditionalCuisineCount,
    SteakhouseCuisineCount,
    PacificislandsCuisineCount,
    EastAsiaCuisineCount,
    BritishCuisineCount,
    CaribbeanCuisineCount,
    FilipinoCuisineCount,
    TurkishCuisineCount,
    BelgianCuisineCount,
    SouthAmericanCuisineCount,
    NorthAfricanCuisineCount,
    SouthAfricanCuisineCount,
    WestAfricanCuisineCount,
    EastAfricanCuisineCount,
    WesternEuropeanCuisineCount,
    PortugueseCuisineCount,
    PeruvianCuisineCount,
    JapaneseCuisineCount,
    PacificIslandsCuisineCount,
    SouthernEuropeCuisineCount,
    AfricanCuisineCount,
    CentralAsiaCuisineCount,
    NorthamericaCuisineCount,
    EuropeanCuisineCount,
    MoroccanCuisineCount,
    AustralianCuisineCount,
    HungarianCuisineCount,
    IranianCuisineCount,
    SoutheastAsiaCuisineCount,
    HawaiianCuisineCount,
    ScandinavianCuisineCount,
    BrazilianCuisineCount,
    IndonesianCuisineCount,
    MongolianCuisineCount,
    RussianCuisineCount,
    SouthAsianCuisineCount,
    BulgarianCuisineCount,
    FusionCuisineCusiineCount,
    IrishCuisineCount,
    GeorgianCuisineCount,
    SouthwestCuisineCount,
    CambodianCuisineCount,
    LatinCuisineCount,
    CubanCuisineCount,
    SoutheastCuisineCount,
    SouthHyphenAfricanCuisineCount,
    JamaicanCuisineCount,
    IsrealiCuisineCount,
    EasteuropeanCuisineCount,
    SingaporeanCuisineCount,
    NordicCuisineCount,
    WestHyphenAfricanCuisineCount,
    NortheastCuisineCount,
    TonganCuisineCount,
    WestafricaCuisineCount,
    ZanzibarianCuisineCount,
    MidwestCuisineCount,
)


class MealkitFeatures(FeatureGenerator):
    def __init__(
        self,
        as_at: date,
        feature_periods: FeaturePeriodsType = None,
        feature_stems: set[str] | list[str] | None = None,
        first_day_of_week: str | None = None,
        use_absolute_periods: bool = False,
    ) -> None:
        super().__init__(
            as_at,
            feature_periods,
            feature_stems,
            first_day_of_week,
            use_absolute_periods,
        )

    FEATURE_CLASSES: set[type[Feature]] = {
        Count,
        # NetSpend,
        # GrossSpend,
        RecencyDays,
        CustomerCount,
        ApproxCustomerCount,
        Quantity,
        MostRecentPurchaseDate,
        EarliestPurchaseDate,
        OrderPeriods,
        OrderCount,
        ApproxOrderCount,
        ProductCount,
        ApproxProductCount,
        RecipeCount,
        ApproxRecipeCount,
        AvgOrder,
        Discount,
        AvgQuantityPerOrder,
        CyclesSinceLastOrder,
        AvgPurchaseCycle,
        Cuisines,
        CuisineCount,
        # All of the following are counts for a specific cuisine
        ItalianCuisineCount,
        FrenchCuisineCount,
        SpanishCuisineCount,
        SrilankanCuisineCount,
        LebaneseCuisineCount,
        GreekCuisineCount,
        VietnameseCuisineCount,
        DanishCuisineCount,
        WesternEuropeCuisineCount,
        FusionCuisineCount,
        DutchCuisineCount,
        KoreanCuisineCount,
        SouthEastAsianCuisineCount,
        MalaysianCuisineCount,
        AmericanCuisineCount,
        AsianCuisineCount,
        IndianCuisineCount,
        GermanCuisineCount,
        CentralAmericaCuisineCount,
        ArgentinianCuisineCount,
        NorthAmericanCuisineCount,
        NewZealandCuisineCount,
        CanadianCuisineCount,
        SwedishCuisineCount,
        EgyptianCuisineCount,
        NorthernEuropeanCuisineCount,
        CajunCuisineCount,
        AustrianCuisineCount,
        MexicanCuisineCount,
        MediterraneanCuisineCount,
        ThaiCuisineCount,
        ChineseCuisineCount,
        LatinAmericanCuisineCount,
        SouthAsiaCuisineCount,
        MiddleEasternCuisineCount,
        TraditionalCuisineCount,
        SteakhouseCuisineCount,
        PacificislandsCuisineCount,
        EastAsiaCuisineCount,
        BritishCuisineCount,
        CaribbeanCuisineCount,
        FilipinoCuisineCount,
        TurkishCuisineCount,
        BelgianCuisineCount,
        SouthAmericanCuisineCount,
        NorthAfricanCuisineCount,
        SouthAfricanCuisineCount,
        WestAfricanCuisineCount,
        EastAfricanCuisineCount,
        WesternEuropeanCuisineCount,
        PortugueseCuisineCount,
        PeruvianCuisineCount,
        JapaneseCuisineCount,
        PacificIslandsCuisineCount,
        SouthernEuropeCuisineCount,
        AfricanCuisineCount,
        CentralAsiaCuisineCount,
        NorthamericaCuisineCount,
        EuropeanCuisineCount,
        MoroccanCuisineCount,
        AustralianCuisineCount,
        HungarianCuisineCount,
        IranianCuisineCount,
        SoutheastAsiaCuisineCount,
        HawaiianCuisineCount,
        ScandinavianCuisineCount,
        BrazilianCuisineCount,
        IndonesianCuisineCount,
        MongolianCuisineCount,
        RussianCuisineCount,
        SouthAsianCuisineCount,
        BulgarianCuisineCount,
        FusionCuisineCusiineCount,
        IrishCuisineCount,
        GeorgianCuisineCount,
        SouthwestCuisineCount,
        CambodianCuisineCount,
        LatinCuisineCount,
        CubanCuisineCount,
        SoutheastCuisineCount,
        SouthHyphenAfricanCuisineCount,
        JamaicanCuisineCount,
        IsrealiCuisineCount,
        EasteuropeanCuisineCount,
        SingaporeanCuisineCount,
        NordicCuisineCount,
        WestHyphenAfricanCuisineCount,
        NortheastCuisineCount,
        TonganCuisineCount,
        WestafricaCuisineCount,
        ZanzibarianCuisineCount,
        MidwestCuisineCount,
    }



"""EarliestPurchaseDate feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.min_feature import Min


class EarliestPurchaseDate(Min):
    def column_expression(self) -> Column:
        return f.to_date(f.col("Timestamp"))

    @property
    def description_subject(self) -> str:
        return "Earliest purchase date"

    def default_value(self) -> Column:
        return f.lit(None)

    @property
    def commentary(self) -> str:
        return (
            "Basically, when was a purchase first made. This could "
            + "be used to determine a whole myriad of useful information"
            + " such as when a customer first shopped, or"
            + "when a customer first bought a particular product. To make "
            + "use of this you would likely need to define "
            + "a large feature period, there isn't much use"
            + "in knowing when someone's earliest purchase was "
            + "in the previous two weeks."
        )



"""RecipeCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class RecipeCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Recipe")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Recipes"

    @property
    def commentary(self) -> str:
        return (
            "The number of recipes. Typically the dataframe supplied "
            + "to this feature will have many recipes for the same order, "
            + "this feature allows you to determine how many distinct recipes "
            + "existed in that collection of recipes."
        )



"""StoreCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class StoreCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Store")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Stores"

    @property
    def commentary(self) -> str:
        return (
            "The number of stores. Typically the dataframe supplied "
            + "to this feature will have many transactions for many baskets for "
            + "many stores, this feature allows you to determine how many "
            + "stores those baskets were purchased in."
        )



"""CyclesSinceLastPurchase feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.feature import DerivedFeature
from .average_purchase_cycle import AvgPurchaseCycle
from jstark.features.recency_days import RecencyDays


class CyclesSinceLastPurchase(DerivedFeature):
    def column_expression(self) -> Column:
        return f.try_divide(
            RecencyDays(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
            AvgPurchaseCycle(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
        )

    @property
    def description_subject(self) -> str:
        return "Cycles since last purchase"

    @property
    def commentary(self) -> str:
        return (
            "Days since last purchase divided by average purchase cycle. "
            + "This may be a predictor of when a customer is "
            + "likely to next buy a particular product."
        )

    def default_value(self) -> Column:
        return f.lit(None)



"""CountIf feature"""

from typing import Callable
import pyspark.sql.functions as f
from pyspark.sql import Column


from jstark.features.feature import BaseFeature


class CountIf(BaseFeature):
    def aggregator(self) -> Callable[[Column], Column]:
        return self.count_if_aggregator

    def column_expression(self) -> Column:
        return f.lit(1)

    def default_value(self) -> Column:
        return f.lit(0)



"""Quantity feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.sum_feature import Sum


class Quantity(Sum):
    def column_expression(self) -> Column:
        return f.col("Quantity")

    @property
    def description_subject(self) -> str:
        return "Sum of Quantity"

    def default_value(self) -> Column:
        return f.lit(0)



"""FirstAndLastDateOfPeriod class

Helper class for figuring out dates relative to a given date
"""

from datetime import date, timedelta
from dateutil.relativedelta import relativedelta


class FirstAndLastDateOfPeriod:
    """Encapsulate all the logic to determine first and last date
    of a period that includes the supplied date
    """

    def __init__(
        self, date_in_period: date, first_day_of_week: str | None = None
    ) -> None:
        self._date_in_period = date_in_period
        self._weekdays = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        if first_day_of_week is None:
            # use what python determines to be the first day of the week
            first_day_of_week = (
                date.today() - timedelta(date.today().weekday())
            ).strftime("%A")
        if first_day_of_week not in self._weekdays:
            raise ValueError(f"first_day_of_week must be one of {self._weekdays}")
        self._first_day_of_week = first_day_of_week

    @property
    def first_date_in_week(self) -> date:
        current_weekday_index = self._weekdays.index(
            self._date_in_period.strftime("%A")
        )
        first_day_index = self._weekdays.index(self._first_day_of_week)
        # Number of days to subtract to get to the first day of this week (may be 0)
        days_to_subtract = (current_weekday_index - first_day_index) % 7
        return self._date_in_period - timedelta(days=days_to_subtract)

    @property
    def last_date_in_week(self) -> date:
        return self.first_date_in_week + timedelta(days=6)

    @property
    def first_date_in_month(self) -> date:
        return date(self._date_in_period.year, self._date_in_period.month, 1)

    @property
    def last_date_in_month(self) -> date:
        return (
            self._date_in_period
            + relativedelta(months=1, day=1)
            - relativedelta(days=1)
        )

    @property
    def first_date_in_quarter(self) -> date:
        match self._date_in_period.month:
            case 1 | 2 | 3:
                return date(self._date_in_period.year, 1, 1)
            case 4 | 5 | 6:
                return date(self._date_in_period.year, 4, 1)
            case 7 | 8 | 9:
                return date(self._date_in_period.year, 7, 1)
            case _:  # all other months:
                return date(self._date_in_period.year, 10, 1)

    @property
    def last_date_in_quarter(self) -> date:
        match self._date_in_period.month:
            case 1 | 2 | 3:
                return date(self._date_in_period.year, 3, 31)
            case 4 | 5 | 6:
                return date(self._date_in_period.year, 6, 30)
            case 7 | 8 | 9:
                return date(self._date_in_period.year, 9, 30)
            case _:  # all other months:
                return date(self._date_in_period.year, 12, 31)

    @property
    def first_date_in_year(self) -> date:
        return date(self._date_in_period.year, 1, 1)

    @property
    def last_date_in_year(self) -> date:
        return date(self._date_in_period.year, 12, 31)



"""ApproxOrderCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.approxdistinctcount_feature import ApproxDistinctCount


class ApproxOrderCount(ApproxDistinctCount):
    def column_expression(self) -> Column:
        return f.col("Order")

    @property
    def description_subject(self) -> str:
        return "Approximate distinct count of Orders"

    @property
    def commentary(self) -> str:
        return (
            "The approximate number of orders. Similar to "
            + f"OrderCount_{self.feature_period.mnemonic} "
            + "except that it uses an approximation algorithm which "
            + "will not be as accurate as "
            + f"OrderCount_{self.feature_period.mnemonic} but will be a lot "
            + 'quicker to compute and in many cases will be "close enough".'
        )



"""AvgPurchaseCycle feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.feature import DerivedFeature
from jstark.features.earliest_purchase_date import EarliestPurchaseDate
from jstark.features.most_recent_purchase_date import MostRecentPurchaseDate
from .basket_count import BasketCount


class AvgPurchaseCycle(DerivedFeature):
    def column_expression(self) -> Column:
        return f.try_divide(
            f.datediff(
                MostRecentPurchaseDate(
                    self.as_at,
                    self.feature_period,
                    first_day_of_week=self._first_day_of_week,
                ).column,
                EarliestPurchaseDate(
                    self.as_at,
                    self.feature_period,
                    first_day_of_week=self._first_day_of_week,
                ).column,
            ),
            BasketCount(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
        )

    @property
    def description_subject(self) -> str:
        return "Average purchase cycle"

    @property
    def commentary(self) -> str:
        return (
            "How often (measured in days) is a purchase made. This "
            + "is very useful to determine how often a customer buys "
            + "a particular product"
        )

    def default_value(self) -> Column:
        return f.lit(None)



"""Discount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.sum_feature import Sum


class Discount(Sum):
    def column_expression(self) -> Column:
        return f.col("Discount")

    @property
    def description_subject(self) -> str:
        return "Sum of Discount"

    @property
    def commentary(self) -> str:
        return "Requires a field called `Discount` in the input data."



"""OrderPeriods feature"""

import functools
import operator

from jstark.features.feature import DerivedFeature

from pyspark.sql import Column
import pyspark.sql.functions as f

from .order_count import OrderCount
from jstark.feature_period import FeaturePeriod


class OrderPeriods(DerivedFeature):
    """OrderPeriods feature"""

    def column_expression(self) -> Column:
        exprs = []
        for period in range(self.feature_period.end, self.feature_period.start + 1):
            exprs.append(
                f.when(
                    OrderCount(
                        as_at=self.as_at,
                        feature_period=FeaturePeriod(
                            self.feature_period.period_unit_of_measure, period, period
                        ),
                        first_day_of_week=self._first_day_of_week,
                    ).column
                    > 0,
                    1,
                ).otherwise(0)
            )
        return functools.reduce(operator.add, exprs)

    def default_value(self) -> Column:
        return f.lit(None)

    @property
    def description_subject(self) -> str:
        return (
            f"Number of {self.feature_period.period_unit_of_measure.name.lower()}s"
            + " in which at least one order was placed"
        )

    @property
    def commentary(self) -> str:
        return (
            f"The number of {self.feature_period.period_unit_of_measure.name.lower()}s "
            + "in which at least one basket was purchased. The value will be in the "
            + f"range 0 to {self.feature_period.start - self.feature_period.end + 1} "
            + f"because {self.feature_period.start - self.feature_period.end + 1} is "
            + f"the number of {self.feature_period.period_unit_of_measure.name.lower()}"
            + f"s between {self.start_date.strftime('%Y-%m-%d')} and"
            + f" {self.end_date.strftime('%Y-%m-%d')}. When grouped by Customer and"
            + " Product this feature is a useful indicator of the frequency of"
            + " which a Customer purchases a Product."
        )

    @property
    def feature_name(self) -> str:
        return (
            f"Order{self.feature_period.period_unit_of_measure.name.title()}s"
            + f"_{self.feature_period.mnemonic}"
        )



from platform import python_version
from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta

import pytest
from packaging import version
from pyspark.sql import DataFrame
from jstark.exceptions import FeaturePeriodEndGreaterThanStartError
from jstark.feature_period import (
    FeaturePeriod,
    PeriodUnitOfMeasure,
    ALL_MONTHS_LAST_YEAR,
    ALL_QUARTERS_LAST_YEAR,
    ALL_MONTHS_THIS_YEAR,
    ALL_QUARTERS_THIS_YEAR,
    TODAY,
    YESTERDAY,
    THIS_WEEK,
    LAST_WEEK,
    THIS_MONTH,
    LAST_MONTH,
    THIS_QUARTER,
    LAST_QUARTER,
    THIS_YEAR,
    LAST_YEAR,
    THIS_MONTH_VS_ONE_YEAR_PRIOR,
    LAST_MONTH_VS_ONE_YEAR_PRIOR,
    THIS_QUARTER_VS_ONE_YEAR_PRIOR,
    LAST_QUARTER_VS_ONE_YEAR_PRIOR,
    LAST_FIVE_YEARS,
)
from jstark.grocery import GroceryFeatures


def test_feature_period_description():
    assert (
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2).description
        == "Between 3 and 2 days ago"
    )


def test_repr():
    assert (
        repr(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2))
        == "FeaturePeriod(period_unit_of_measure=PeriodUnitOfMeasure.DAY, start=3, end=2)"  # noqa: E501
    )


def test_eq():
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) == FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 3, 2
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 3, 3
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.WEEK, 3, 2
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 4, 2
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 3, 3
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 3, 1
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != FeaturePeriod(
        PeriodUnitOfMeasure.DAY, 2, 2
    )
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2) != "SomeOtherObject"


def test_str():
    assert (
        str(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) == "Between 3 and 2 days ago"
    )


def test_hash():
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) == hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 3)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.WEEK, 3, 2)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 4, 2)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 3)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 1)
    )
    assert hash(FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)) != hash(
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 2)
    )


def test_feature_period_start_is_immutable():
    fp = FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)
    with pytest.raises(AttributeError) as exc_info:
        fp.start = 10  # type: ignore
    if version.parse(python_version()) < version.parse("3.11"):
        assert "can't set attribute" in str(exc_info.value)
    else:
        assert (
            str(exc_info.value)
            == "property 'start' of 'FeaturePeriod' object has no setter"
        )


def test_feature_period_end_is_immutable():
    fp = FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)
    with pytest.raises(AttributeError) as exc_info:
        fp.end = 10  # type: ignore
    if version.parse(python_version()) < version.parse("3.11"):
        assert "can't set attribute" in str(exc_info.value)
    else:
        assert (
            str(exc_info.value)
            == "property 'end' of 'FeaturePeriod' object has no setter"
        )


def test_feature_period_periodunitofmeasure_is_immutable():
    fp = FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2)
    with pytest.raises(AttributeError) as exc_info:
        fp.period_unit_of_measure = PeriodUnitOfMeasure.YEAR  # type: ignore
    if version.parse(python_version()) < version.parse("3.11"):
        assert "can't set attribute" in str(exc_info.value)
    else:
        assert str(exc_info.value) == (
            "property 'period_unit_of_measure' of "
            + "'FeaturePeriod' object has no setter"
        )


def test_feature_period_end_greater_than_start_raises_exception():
    with pytest.raises(FeaturePeriodEndGreaterThanStartError) as exc_info:
        FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 3)
    assert (
        str(exc_info.value)
        == "End of the feature period (3) cannot be "
        + "before the start of the feature period (2)"
    )


def test_feature_period_code_for_day():
    assert FeaturePeriod(PeriodUnitOfMeasure.DAY, 3, 2).mnemonic == "3d2"


def test_feature_period_code_for_week():
    assert FeaturePeriod(PeriodUnitOfMeasure.WEEK, 3, 2).mnemonic == "3w2"


def test_feature_period_code_for_month():
    assert FeaturePeriod(PeriodUnitOfMeasure.MONTH, 3, 2).mnemonic == "3m2"


def test_feature_period_code_for_quarter():
    assert FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 3, 2).mnemonic == "3q2"


def test_feature_period_code_for_year():
    assert FeaturePeriod(PeriodUnitOfMeasure.YEAR, 3, 2).mnemonic == "3y2"


def test_argument_of_wrong_types_raises():
    with pytest.raises(TypeError) as exc_info:
        FeaturePeriod("YEAR", 3, 2)
    assert str(exc_info.value) == (
        "period_unit_of_measure needs to be of type "
        + "PeriodUnitOfMeasure, not <class 'str'>"
    )


def test_all_months_last_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**ALL_MONTHS_LAST_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    last_year = date.today().year - 1
    expected_start_dates = {date(last_year, i, 1) for i in range(1, 13)}
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == expected_start_dates


def test_all_months_this_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**ALL_MONTHS_THIS_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    expected_start_dates = {
        date(date.today().year, i, 1) for i in range(1, date.today().month + 1)
    }
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == expected_start_dates


def test_all_quarters_last_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**ALL_QUARTERS_LAST_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    last_year = date.today().year - 1
    expected_start_dates = {
        date(last_year, 1, 1),
        date(last_year, 4, 1),
        date(last_year, 7, 1),
        date(last_year, 10, 1),
    }
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == expected_start_dates


def test_all_quarters_this_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**ALL_QUARTERS_THIS_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    match date.today().month:
        case 1 | 2 | 3:
            first_month_of_quarters = [1]
        case 4 | 5 | 6:
            first_month_of_quarters = [1, 4]
        case 7 | 8 | 9:
            first_month_of_quarters = [1, 4, 7]
        case _:  # all other months:
            first_month_of_quarters = [1, 4, 7, 10]
    expected_start_dates = {
        date(date.today().year, i, 1) for i in first_month_of_quarters
    }
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == expected_start_dates


def test_today(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**TODAY, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today()}


def test_yesterday(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**YESTERDAY, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today() - timedelta(days=1)}


def test_this_week(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**THIS_WEEK, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    today_day_of_the_week = date.today().weekday()
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today() - timedelta(days=today_day_of_the_week)}


def test_last_week(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_WEEK, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    today_day_of_the_week = date.today().weekday()
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today() - timedelta(days=today_day_of_the_week) - timedelta(days=7)}


def test_this_month(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**THIS_MONTH, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today().replace(day=1)}


def test_last_month(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_MONTH, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today().replace(day=1) - relativedelta(months=1)}


def test_this_quarter(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**THIS_QUARTER, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    match date.today().month:
        case 1 | 2 | 3:
            first_day_of_quarter = date.today().replace(day=1).replace(month=1)
        case 4 | 5 | 6:
            first_day_of_quarter = date.today().replace(day=1).replace(month=4)
        case 7 | 8 | 9:
            first_day_of_quarter = date.today().replace(day=1).replace(month=7)
        case _:  # all other months:
            first_day_of_quarter = date.today().replace(day=1).replace(month=10)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {first_day_of_quarter}


def test_last_quarter(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_QUARTER, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    match date.today().month:
        case 1 | 2 | 3:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=1
            ) - relativedelta(months=3)
        case 4 | 5 | 6:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=4
            ) - relativedelta(months=3)
        case 7 | 8 | 9:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=7
            ) - relativedelta(months=3)
        case _:  # all other months:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=10
            ) - relativedelta(months=3)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {first_day_of_quarter}


def test_this_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**THIS_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today().replace(day=1).replace(month=1)}


def test_last_year(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_YEAR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {date.today().replace(day=1).replace(month=1) - relativedelta(years=1)}


def test_this_month_vs_one_year_prior(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**THIS_MONTH_VS_ONE_YEAR_PRIOR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {
        date.today().replace(day=1),
        date.today().replace(day=1) - relativedelta(years=1),
    }


def test_last_month_vs_one_year_prior(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_MONTH_VS_ONE_YEAR_PRIOR, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {
        date.today().replace(day=1) - relativedelta(months=1),
        date.today().replace(day=1) - relativedelta(months=1) - relativedelta(years=1),
    }


def test_this_quarter_vs_one_year_prior(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(
        **THIS_QUARTER_VS_ONE_YEAR_PRIOR,  # type: ignore[arg-type]
        feature_stems={"BasketCount"},
    )
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    match date.today().month:
        case 1 | 2 | 3:
            first_day_of_quarter = date.today().replace(day=1).replace(month=1)
        case 4 | 5 | 6:
            first_day_of_quarter = date.today().replace(day=1).replace(month=4)
        case 7 | 8 | 9:
            first_day_of_quarter = date.today().replace(day=1).replace(month=7)
        case _:  # all other months:
            first_day_of_quarter = date.today().replace(day=1).replace(month=10)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {
        first_day_of_quarter,
        first_day_of_quarter - relativedelta(years=1),
    }


def test_last_quarter_vs_one_year_prior(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(
        **LAST_QUARTER_VS_ONE_YEAR_PRIOR,  # type: ignore[arg-type]
        feature_stems={"BasketCount"},
    )
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    match date.today().month:
        case 1 | 2 | 3:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=1
            ) - relativedelta(months=3)
        case 4 | 5 | 6:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=4
            ) - relativedelta(months=3)
        case 7 | 8 | 9:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=7
            ) - relativedelta(months=3)
        case _:  # all other months:
            first_day_of_quarter = date.today().replace(day=1).replace(
                month=10
            ) - relativedelta(months=3)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {
        first_day_of_quarter,
        first_day_of_quarter - relativedelta(years=1),
    }


def test_last_five_years(luke_and_leia_purchases: DataFrame):
    gf = GroceryFeatures(**LAST_FIVE_YEARS, feature_stems={"BasketCount"})  # type: ignore[arg-type]
    df = luke_and_leia_purchases.groupBy().agg(*gf.features)
    assert {
        datetime.strptime(c.metadata["start-date"], "%Y-%m-%d").date()
        for c in df.schema
    } == {
        date.today().replace(day=1).replace(month=1) - relativedelta(years=1),
        date.today().replace(day=1).replace(month=1) - relativedelta(years=2),
        date.today().replace(day=1).replace(month=1) - relativedelta(years=3),
        date.today().replace(day=1).replace(month=1) - relativedelta(years=4),
        date.today().replace(day=1).replace(month=1) - relativedelta(years=5),
    }



"""jstark exceptions"""

from jstark.period_unit_of_measure import PeriodUnitOfMeasure


class FeaturePeriodMnemonicIsInvalid(Exception):
    """FeaturePeriod mnemonic doesn't match X[dqmqy]Y"""

    def __str__(self) -> str:
        return (
            "FeaturePeriod mnemonic must be an integer followed by a letter "
            + f"from {[e.value for e in PeriodUnitOfMeasure]} followed by an integer"
        )


class FeaturePeriodEndGreaterThanStartError(Exception):
    """Exception indicating end of a feature period cannot be before the start"""

    def __init__(self, start: int, end: int, *args: object) -> None:
        super().__init__(*args)
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return (
            f"End of the feature period ({self.end}) cannot be "
            + f"before the start of the feature period ({self.start})"
        )


class AsAtIsNotADate(Exception):
    """Exception indicating a value must be of type date"""

    def __str__(self) -> str:
        return "as_at value must be of type date"



"""CuisineCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount
from jstark.features.count_if import CountIf


class CuisineCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Cuisine")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Cuisines"

    @property
    def commentary(self) -> str:
        return (
            "The number of cuisines. Typically the dataframe supplied "
            + "to this feature will have many recipes for the same cuisine, "
            + "this feature allows you to determine how many distinct cuisines "
            + "have been ordered."
        )


class CuisineCountIf(CountIf):
    CUISINE_NAME: str = ""

    def column_expression(self) -> Column:
        # Compare lowercase cuisine to class's CUISINE_NAME (also lowercase)
        return f.lower(f.col("Cuisine")) == self.CUISINE_NAME.lower()

    @property
    def description_subject(self) -> str:
        return f"Count of {self.CUISINE_NAME.capitalize()} recipes"

    @property
    def commentary(self) -> str:
        name = self.CUISINE_NAME.capitalize()
        return (
            f"The number of {name} recipes. Typically the "
            + "dataframe supplied to this feature will have "
            + "many recipes for the same cuisine, this feature "
            + f"allows you to determine how many {name} "
            + "recipes have been ordered."
        )


# This list of classes was based upon all the observed cuisines in
# all recipes from a well-known mealkit provider in 2025
class ItalianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "italian"


class FrenchCuisineCount(CuisineCountIf):
    CUISINE_NAME = "french"


class SpanishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "spanish"


class SrilankanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "srilankan"


class LebaneseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "lebanese"


class GreekCuisineCount(CuisineCountIf):
    CUISINE_NAME = "greek"


class VietnameseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "vietnamese"


class DanishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "danish"


class WesternEuropeCuisineCount(CuisineCountIf):
    CUISINE_NAME = "westerneurope"


class FusionCuisineCount(CuisineCountIf):
    CUISINE_NAME = "fusion"


class DutchCuisineCount(CuisineCountIf):
    CUISINE_NAME = "dutch"


class KoreanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "korean"


class SouthEastAsianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "southeast-asian"


class MalaysianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "malay"


class AmericanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "american"


class AsianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "asian"


class IndianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "indian"


class GermanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "german"


class CentralAmericaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "central america"


class ArgentinianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "argentinian"


class NorthAmericanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "north american"


class NewZealandCuisineCount(CuisineCountIf):
    CUISINE_NAME = "new zealand"


class CanadianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "canadian"


class SwedishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "swedish"


class EgyptianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "egyptian"


class NorthernEuropeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "northern europe"


class CajunCuisineCount(CuisineCountIf):
    CUISINE_NAME = "cajun"


class AustrianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "austrian"


class MexicanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "mexican"


class MediterraneanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "mediterranean"


class ThaiCuisineCount(CuisineCountIf):
    CUISINE_NAME = "thai"


class ChineseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "chinese"


class LatinAmericanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "latin american"


class SouthAsiaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "south asia"


class MiddleEasternCuisineCount(CuisineCountIf):
    CUISINE_NAME = "middleeastern"


class TraditionalCuisineCount(CuisineCountIf):
    CUISINE_NAME = "traditional"


class SteakhouseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "steakhouse"


class PacificislandsCuisineCount(CuisineCountIf):
    CUISINE_NAME = "pacificislands"


class EastAsiaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "east asia"


class BritishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "british"


class CaribbeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "caribbean"


class FilipinoCuisineCount(CuisineCountIf):
    CUISINE_NAME = "filipino"


class TurkishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "turkish"


class BelgianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "belgian"


class SouthAmericanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "south american"


class NorthAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "north african"


class SouthAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "south african"


class WestAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "west african"


class EastAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "east african"


class WesternEuropeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "western-european"


class PortugueseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "portuguese"


class PeruvianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "peruvian"


class JapaneseCuisineCount(CuisineCountIf):
    CUISINE_NAME = "japanese"


class PacificIslandsCuisineCount(CuisineCountIf):
    CUISINE_NAME = "pacific-islands"


class SouthernEuropeCuisineCount(CuisineCountIf):
    CUISINE_NAME = "southern europe"


class AfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "african"


class CentralAsiaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "centralasia"


class NorthamericaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "northamerica"


class EuropeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "european"


class MoroccanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "moroccan"


class AustralianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "australian"


class HungarianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "hungarian"


class IranianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "iranian"


class SoutheastAsiaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "southeast asia"


class HawaiianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "hawaiian"


class ScandinavianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "scandinavian"


class BrazilianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "brazilian"


class IndonesianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "indonesian"


class MongolianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "mongolian"


class RussianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "russian"


class SouthAsianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "south-asian"


class BulgarianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "bulgarian"


class FusionCuisineCusiineCount(CuisineCountIf):
    CUISINE_NAME = "fusion-cuisine"


class IrishCuisineCount(CuisineCountIf):
    CUISINE_NAME = "irish"


class GeorgianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "georgian"


class SouthwestCuisineCount(CuisineCountIf):
    CUISINE_NAME = "southwest"


class CambodianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "cambodian"


class LatinCuisineCount(CuisineCountIf):
    CUISINE_NAME = "latin"


class CubanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "cuban"


class SoutheastCuisineCount(CuisineCountIf):
    CUISINE_NAME = "southeast"


class SouthHyphenAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "south-african"


class JamaicanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "jamaican"


class IsrealiCuisineCount(CuisineCountIf):
    CUISINE_NAME = "israeli"


class EasteuropeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "easteuropean"


class SingaporeanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "singaporean"


class NordicCuisineCount(CuisineCountIf):
    CUISINE_NAME = "nordic"


class WestHyphenAfricanCuisineCount(CuisineCountIf):
    CUISINE_NAME = "west-african"


class NortheastCuisineCount(CuisineCountIf):
    CUISINE_NAME = "northeast"


class TonganCuisineCount(CuisineCountIf):
    CUISINE_NAME = "tongan"


class WestafricaCuisineCount(CuisineCountIf):
    CUISINE_NAME = "westafrica"


class ZanzibarianCuisineCount(CuisineCountIf):
    CUISINE_NAME = "zanzibarian"


class MidwestCuisineCount(CuisineCountIf):
    CUISINE_NAME = "midwest"



"""MostRecentPurchaseDate feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.max_feature import Max


class MostRecentPurchaseDate(Max):
    def column_expression(self) -> Column:
        return f.to_date(f.col("Timestamp"))

    @property
    def description_subject(self) -> str:
        return "Most recent purchase date"

    def default_value(self) -> Column:
        return f.lit(None)

    @property
    def commentary(self) -> str:
        return (
            "It is useful to be able to see when something was most recently "
            + "purchased.For example, grouping by Store and filtering where "
            + "MostRecentPurchaseDate is more than 2 days ago could be a useful "
            + "indicator of things which might not be available for purchase."
            + "Also note that this is very similar to "
            + f"RecencyDays_{self.feature_period.mnemonic} so consider which of these "
            + "features is most useful to you."
        )



"""MaxGrossPrice feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.max_feature import Max


class MaxGrossPrice(Max):
    def column_expression(self) -> Column:
        return f.try_divide(f.col("GrossSpend"), f.col("Quantity"))

    @property
    def description_subject(self) -> str:
        return "Maximum of (GrossSpend / Quantity)"

    @property
    def commentary(self) -> str:
        return (
            "The gross price is calculated as the gross "
            + "spend divided by how many were bought."
        )
